"""
Utilities for working with annotations in
`Label Studio <https://labelstud.io>`_.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from copy import deepcopy
from datetime import datetime
import itertools
import json
import logging
from packaging import version
import random
import string
import webbrowser

from bson import ObjectId
import numpy as np

import fiftyone as fo
import fiftyone.core.labels as fol
import fiftyone.core.media as fom
import fiftyone.core.utils as fou
import fiftyone.utils.annotations as foua

ls = fou.lazy_import(
    "label_studio_sdk",
    callback=lambda: fou.ensure_import("label_studio_sdk>=0.0.13"),
)


logger = logging.getLogger(__name__)


class LabelStudioBackendConfig(foua.AnnotationBackendConfig):
    """Class for configuring :class:`LabelStudioBackend` instances.

    Args:
        name: the name of the backend
        label_schema: a dictionary containing the description of label fields,
            classes and attribute to annotate
        media_field ("filepath"): string field name containing the paths to
            media files on disk to upload
        url: the URL of the Label Studio server
        api_key: the API key to use for authentication
        project_name: the name of the project to use on the Label Studio server
    """

    def __init__(
        self,
        name,
        label_schema,
        media_field="filepath",
        url=None,
        api_key=None,
        project_name=None,
        **kwargs,
    ):
        super().__init__(name, label_schema, media_field=media_field, **kwargs)

        self.url = url
        self.project_name = project_name

        self._api_key = api_key

    @property
    def api_key(self):
        return self._api_key

    @api_key.setter
    def api_key(self, value):
        self._api_key = value


class LabelStudioBackend(foua.AnnotationBackend):
    """Class for interacting with the Label Studio annotation backend."""

    @property
    def supported_media_types(self):
        return [fom.IMAGE]

    @property
    def supported_label_types(self):
        return [
            "classification",
            "detection",
            "detections",
            "instance",
            "instances",
            "polyline",
            "polylines",
            "polygon",
            "polygons",
            "keypoint",
            "keypoints",
            "segmentation",
            "scalar",
        ]

    @property
    def supported_scalar_types(self):
        return []

    @property
    def supported_attr_types(self):
        return []

    @property
    def supports_keyframes(self):
        return False

    @property
    def supports_video_sample_fields(self):
        return False

    @property
    def requires_label_schema(self):
        return True

    def _connect_to_api(self):
        return LabelStudioAnnotationAPI(
            url=self.config.url, api_key=self.config.api_key
        )

    def upload_annotations(self, samples, launch_editor=False):
        api = self.connect_to_api()

        logger.info("Uploading media to Label Studio...")
        results = api.upload_samples(samples, self)
        logger.info("Upload complete")

        if launch_editor:
            results.launch_editor()

        return results

    def download_annotations(self, results):
        api = self.connect_to_api()

        logger.info("Downloading labels from Label Studio...")
        annotations = api.download_annotations(results)
        logger.info("Download complete")

        return annotations


class LabelStudioAnnotationAPI(foua.AnnotationAPI):
    """A class to upload tasks and predictions to and fetch annotations from
    Label Studio.

    On initialization, the class will check if the server is reachable.
    """

    def __init__(self, url, api_key):
        self.url = url
        self._api_key = api_key
        self.backend = "labelstudio"
        self._min_server_version = "1.5.0"

        self._setup()

    def _setup(self):
        if self._api_key is None:
            self._api_key = self._prompt_api_key(self.backend)

        self._client = ls.Client(self.url, self._api_key)
        self._client.check_connection()
        self._verify_server_version()

    def _verify_server_version(self):
        server_version = self._client.make_request(
            "GET", "/api/version"
        ).json()["release"]
        if not version.parse(server_version) >= version.parse(
            self._min_server_version
        ):
            raise ValueError(
                "Current Label Studio integration is only compatible with "
                "version>=%s" % self._min_server_version
            )

    def _init_project(self, config, samples):
        """Creates a new project on Label Studio.

        If project_name is not set in the configs, it will be generated.
        If project_name exists on the server, a timestamp will be added
        to the project name.

        Args:
            config: a :class:`LabelStudioBackendConfig`
            samples: a :class:`fiftyone.core.collections.SampleCollection`

        Returns:
            a ``label_studio_sdk.Project``
        """
        project_name = deepcopy(config.project_name)
        label_schema = deepcopy(config.label_schema)

        if project_name is None:
            _dataset_name = samples._root_dataset.name.replace(" ", "_")
            project_name = "FiftyOne_%s" % _dataset_name

        # if project name take, add timestamp
        projects = self._client.list_projects()
        for one in projects:
            if one.params["title"] == project_name:
                time_str = str(int(datetime.timestamp(datetime.now())))
                project_name += "_%s" % time_str
                break

        # generate label config
        assert len(label_schema) == 1
        _, label_info = label_schema.popitem()
        label_config = generate_labeling_config(
            media=samples.media_type,
            label_type=label_info["type"],
            labels=label_info["classes"],
        )

        project = self._client.start_project(
            title=project_name, label_config=label_config
        )
        return project

    def _prepare_tasks(self, samples, label_schema, media_field):
        """Prepares Label Studio tasks for the given data."""
        samples.compute_metadata()

        ids, mime_types, filepaths = samples.values(
            ["id", "metadata.mime_type", media_field]
        )

        tasks = [
            {
                "source_id": _id,
                "media_type": "image",
                "mime_type": _mime_type,
                "image": _filepath,
            }
            for _id, _mime_type, _filepath in zip(ids, mime_types, filepaths)
        ]

        predictions, id_map = {}, {}
        for label_field, label_info in label_schema.items():
            if label_info["existing_field"]:
                predictions[label_field] = {
                    smp.id: export_label_to_label_studio(
                        smp[label_field],
                        full_result={
                            "from_name": "label",
                            "to_name": "image",
                            "original_width": smp.metadata["width"],
                            "original_height": smp.metadata["height"],
                            "image_rotation": getattr(smp, "rotation", 0),
                        },
                    )
                    for smp in samples.select_fields(label_field)
                }
                id_map[label_field] = {
                    smp.id: _get_label_ids(smp[label_field])
                    for smp in samples.select_fields(label_field)
                }

        return tasks, predictions, id_map

    def _upload_tasks(self, project, tasks, predictions=None):
        """Uploads files to Label Studio and registers them as tasks.

        Args:
            project: a ``label_studio_sdk.Project``
            tasks: a list of task dicts
            predictions (None): optional predictions to upload

        Returns:
            a dict mapping ``task_id`` to ``sample_id``
        """
        files = [
            (
                one["source_id"],
                (
                    one["source_id"],
                    open(one[one["media_type"]], "rb"),
                    one["mime_type"],
                ),
            )
            for one in tasks
        ]

        # upload files first and get their upload ids
        upload_resp = self._client.make_request(
            "POST",
            f"/api/projects/{project.id}/import",
            params={"commit_to_project": True},
            files=files,
        )

        # create tasks out of the uploaded files
        payload = json.dumps(
            {
                "file_upload_ids": upload_resp.json()["file_upload_ids"],
                "files_as_tasks_list": False,
            }
        )
        self._client.headers.update({"Content-Type": "application/json"})
        self._client.make_request(
            "POST", f"/api/projects/{project.id}/reimport", data=payload
        )

        # get uploaded task ids
        uploaded_ids = project.get_tasks(only_ids=True)[-len(files) :]
        uploaded_tasks = {
            i: t["source_id"] for i, t in zip(uploaded_ids, tasks)
        }

        # upload predictions if given
        if predictions:
            source2task = {v: k for k, v in uploaded_tasks.items()}
            for _, label_predictions in predictions.items():
                ls_predictions = [
                    {
                        "task": source2task[smp_id],
                        "result": pred,
                    }
                    for smp_id, pred in label_predictions.items()
                ]
                project.create_predictions(ls_predictions)

        return uploaded_tasks

    @staticmethod
    def _get_matched_labeled_tasks(project, task_ids):
        matched_tasks = project.get_tasks(selected_ids=task_ids)

        def task_filter(x):
            return x["is_labeled"] or bool(x.get("predictions"))

        return list(filter(task_filter, matched_tasks))

    def _import_annotations(self, tasks, task_map, label_type):
        results = {}
        for t in tasks:
            # convert latest annotation results
            if t["is_labeled"]:
                annotations = t.get("annotations", [])
            else:
                annotations = t.get("predictions", [])

            latest_annotation = (
                annotations[-1]
                if len(annotations) == 0
                else sorted(annotations, key=lambda x: x["updated_at"])[-1]
            )
            if label_type == "keypoints":
                labels = import_label_studio_annotation(
                    latest_annotation["result"]
                )
            else:
                labels = [
                    import_label_studio_annotation(r)
                    for r in latest_annotation.get("result", [])
                ]

            # add to dict
            if labels:
                label_ids = (
                    {l.id: l for l in labels}
                    if not isinstance(labels[0], fol.Regression)
                    else labels[0]
                )
                sample_id = task_map[t["id"]]
                results[sample_id] = label_ids

        return results

    def _export_to_label_studio(self, labels, label_type):
        if _LABEL_TYPES[label_type]["multiple"] is None:
            return export_label_to_label_studio(labels)

        return [export_label_to_label_studio(l) for l in labels]

    def upload_samples(self, samples, backend):
        """Uploads the given samples to Label Studio according to the given
        backend's annotation and server configuration.

        Args:
            samples: a :class:`fiftyone.core.collections.SampleCollection`
            backend: a :class:`LabelStudioBackend` to use to perform the upload

        Returns:
            a :class:`LabelStudioAnnotationResults`
        """
        config = backend.config

        project = self._init_project(config, samples)

        tasks, predictions, id_map = self._prepare_tasks(
            samples,
            config.label_schema,
            config.media_field,
        )
        uploaded_tasks = self._upload_tasks(project, tasks, predictions)

        return LabelStudioAnnotationResults(
            samples,
            config,
            id_map=id_map,
            project_id=project.id,
            uploaded_tasks=uploaded_tasks,
            backend=backend,
        )

    def download_annotations(self, results):
        """Downloads the annotations from the Label Studio server for the given
        results instance and parses them into the appropriate FiftyOne types.

        Args:
            results: a :class:`LabelStudioAnnotationResults`

        Returns:
            the annotations dict
        """
        project = self._client.get_project(results.project_id)
        labeled_tasks = self._get_matched_labeled_tasks(
            project, list(results.uploaded_tasks.keys())
        )
        annotations = {}
        for label_field, label_info in results.config.label_schema.items():
            return_type = foua._RETURN_TYPES_MAP[label_info["type"]]
            labels = self._import_annotations(
                labeled_tasks, results.uploaded_tasks, return_type
            )
            annotations.update({label_field: {return_type: labels}})

        return annotations

    def upload_predictions(self, project, tasks, sample_labels, label_type):
        """Uploads the given predictions to an existing Label Studio project.

        Args:
            project: a ``label_studio_sdk.Project``
            tasks: a list of task dicts
            sample_labels: a list or list of lists of
                :class:`fiftyone.core.labels.Label` instances
            label_type: the label type string
        """
        for task, labels in zip(tasks, sample_labels):
            predictions = self._export_to_label_studio(labels, label_type)
            project.create_prediction(task, predictions)

    def delete_tasks(self, task_ids):
        """Deletes the given tasks from Label Studio.

        Args:
            task_ids: list of task ids
        """
        for t_id in task_ids:
            self._client.make_request(
                "DELETE",
                f"/api/tasks/{t_id}",
            )

    def delete_project(self, project_id):
        """Deletes the project from Label Studio.

        Args:
            project_id: project id
        """
        self._client.make_request(
            "DELETE",
            f"/api/projects/{project_id}",
        )


class LabelStudioAnnotationResults(foua.AnnotationResults):
    """Class that stores all relevant information needed to monitor the
    progress of an annotation run sent to Label Studio and download the
    results.
    """

    def __init__(
        self, samples, config, id_map, project_id, uploaded_tasks, backend=None
    ):
        super().__init__(samples, config, id_map=id_map, backend=backend)
        self.project_id = project_id
        self.uploaded_tasks = uploaded_tasks

    def load_credentials(self, url=None, api_key=None):
        """Load the Label Studio credentials from the given keyword arguments
        or the FiftyOne annotation config.

        Args:
            url (None): the url of the Label Studio server
            api_key (None): the Label Studio API key
        """
        self._load_config_parameters(url=url, api_key=api_key)

    def _load_config_parameters(self, **kwargs):
        config = self.config
        parameters = fo.annotation_config.backends.get(config.name, {})

        for name, value in kwargs.items():
            if value is None:
                value = parameters.get(name, None)

            if value is not None:
                setattr(config, name, value)

    def launch_editor(self):
        """Open a Label Studio tab in browser."""
        project_url = f"{self.config.url}/projects/{self.project_id}"
        logger.info("Launching editor at '%s'...", project_url)
        webbrowser.open_new_tab(project_url)

    def cleanup(self):
        if self.project_id is not None:
            api = self.backend.connect_to_api()
            api.delete_tasks(self.uploaded_tasks)
            api.delete_project(self.project_id)

    @classmethod
    def _from_dict(cls, d, samples, config):
        # int keys were serialized as strings...
        uploaded_tasks = {
            int(task_id): source_id
            for task_id, source_id in d["uploaded_tasks"].items()
        }

        return cls(
            samples,
            config,
            d["id_map"],
            d["project_id"],
            uploaded_tasks,
        )


def generate_labeling_config(media, label_type, labels=None):
    """Generates a labeling config for a Label Studio project.

    Args:
        media: The media type to label
        label_type: The type of labels to use
        labels (None): the labels to use

    Returns:
        a labeling config
    """
    etree = fou.lazy_import(
        "lxml.etree", callback=lambda: fou.ensure_import("lxml.etree")
    )

    assert media in ["image", "video"]
    assert (
        label_type in _LABEL_TYPES.keys()
        or label_type in foua._RETURN_TYPES_MAP.keys()
    )

    # root view and media view
    root = etree.Element("View")
    etree.SubElement(root, media.capitalize(), name=media, value=f"${media}")

    # labels view
    parent_tag, child_tag, tag_kwargs = _ls_tags_from_type(label_type)
    # parent_name = child_tag.lower() if child_tag else parent_tag.lower()
    label_view = etree.SubElement(
        root, parent_tag, name="label", toName=media, **tag_kwargs
    )
    if labels:
        for one in labels:
            etree.SubElement(label_view, child_tag, value=one)

    config_str = etree.tostring(root, pretty_print=True).decode()
    return config_str


def import_label_studio_annotation(result):
    """Imports an annotation from Label Studio.

    Args:
        result: the annotation result from Label Studio

    Returns:
        a :class:`fiftyone.core.labels.Label`
    """
    # TODO link keypoints by parent id
    # TODO handle multiple classes for segmentation
    if isinstance(result, dict):
        ls_type = result["type"]
    elif isinstance(result, list):
        ls_type = result[0]["type"]
    else:
        raise TypeError("Result type %s is not understood" % type(result))

    if ls_type == "choices":
        label = _from_choices(result)
    elif ls_type == "rectanglelabels":
        label = _from_rectanglelabels(result)
    elif ls_type == "polygonlabels":
        label = _from_polygonlabels(result)
    elif ls_type == "keypointlabels":
        label = _from_keypointlabels(result)
    elif ls_type == "brushlabels":
        label = _from_brushlabels(result)
    elif ls_type == "number":
        label = _from_number(result)
    else:
        raise ValueError("Unable to import %s from Label Studio" % ls_type)

    try:
        label_id = result["id"]
        ObjectId(label_id)  # verify that ID is valid
        label.id = label_id
    except:
        pass

    return label


def _update_dict(src_dict, update_dict):
    new = deepcopy(src_dict)
    new.update(update_dict)
    return new


def export_label_to_label_studio(label, full_result=None):
    """Exports a label to the Label Studio format.

    Args:
        label: a :class:`fiftyone.core.labels.Label` or list of
            :class:`fiftyone.core.labels.Label` instances
        full_result (None): if non-empty, return the full Label Studio result

    Returns:
        a dictionary or a list in Label Studio format
    """
    # TODO model version and model score
    if label is None:
        result_value = {}
        ls_type = None
        ids = []
    elif _check_type(label, fol.Classification, fol.Classifications):
        result_value, ls_type, ids = _to_classification(label)
    elif _check_type(label, fol.Detection, fol.Detections):
        result_value, ls_type, ids = _to_detection(label)
    elif _check_type(label, fol.Polyline, fol.Polylines):
        result_value, ls_type, ids = _to_polyline(label)
    elif _check_type(label, fol.Keypoint, fol.Keypoints):
        result_value, ls_type, ids = _to_keypoint(label)
    elif isinstance(label, fol.Segmentation):
        result_value, ls_type, ids = _to_segmentation(label)
    elif isinstance(label, fol.Regression):
        result_value = {"number": label.value}
        ls_type = "number"
        ids = label.id
    else:
        raise ValueError("Label type %s is not supported" % type(label))

    if full_result:
        # return full LS result
        if not isinstance(result_value, (list, tuple)):
            result_value = [result_value]

        return [
            _update_dict(
                full_result,
                dict(value=r, type=ls_type, id=i),
            )
            for r, i in zip(result_value, ids)
        ]

    return result_value


def _generate_prediction_id(n=10):
    return "".join(random.choices(string.ascii_letters, k=n))


def _to_classification(label):
    ls_type = "choices"

    if isinstance(label, list):
        return (
            {ls_type: [l.label for l in label]},
            ls_type,
            [l.id for l in label],
        )

    if isinstance(label, fol.Classifications):
        return (
            {ls_type: [l.label for l in label.classifications]},
            ls_type,
            [l.id for l in label.classifications],
        )

    return {ls_type: [label.label]}, ls_type, label.id


def _to_detection(label):
    ls_type = "rectanglelabels"

    if isinstance(label, list):
        return (
            [_to_detection(l)[0] for l in label],
            ls_type,
            [l.id for l in label],
        )

    if isinstance(label, fol.Detections):
        return (
            [_to_detection(l)[0] for l in label.detections],
            ls_type,
            [l.id for l in label.detections],
        )

    box = _denormalize_values(label.bounding_box)
    result = {
        "x": box[0],
        "y": box[1],
        "width": box[2],
        "height": box[3],
        "rotation": getattr(label, "rotation", 0),
        "rectanglelabels": [label.label],
    }
    return result, ls_type, label.id


def _to_polyline(label):
    ls_type = "polygonlabels"

    if isinstance(label, list):
        return (
            [_to_polyline(l)[0] for l in label],
            ls_type,
            [l.id for l in label],
        )

    if isinstance(label, fol.Polylines):
        return (
            [_to_polyline(l)[0] for l in label.polylines],
            ls_type,
            [l.id for l in label.polylines],
        )

    result = {
        "points": _denormalize_values(label.points[0]),
        "polygonlabels": [label.label],
    }
    return result, ls_type, label.id


def _to_keypoint(label):
    ls_type = "keypointlabels"

    if isinstance(label, list):
        return (
            sum([_to_keypoint(l)[0] for l in label], []),
            ls_type,
            [l.id for l in label],
        )

    if isinstance(label, fol.Keypoints):
        return (
            sum([_to_keypoint(l)[0] for l in label.keypoints], []),
            ls_type,
            [l.id for l in label.keypoints],
        )

    points = _denormalize_values(label.points)
    results = [
        {
            "x": p[0],
            "y": p[1],
            "width": getattr(label, "width", 0.34),
            "keypointlabels": [label.label],
        }
        for p in points
    ]
    return results, ls_type, label.id


def _to_segmentation(label):
    brush = fou.lazy_import(
        "label_studio_converter.brush",
        callback=lambda: fou.ensure_import("label_studio_converter.brush"),
    )
    rle = brush.mask2rle(label.mask)
    result = {"format": "rle", "rle": rle, "brushlabels": [label.label]}
    return result, "brushlabels", label.id


def _from_choices(result):
    label_values = result["value"]["choices"]
    if len(label_values) == 1:
        return fol.Classification(label=label_values[0])

    # multi-label classification
    return [fol.Classification(label=l) for l in label_values]


def _from_rectanglelabels(result):
    ls_box = [
        result["value"]["x"],
        result["value"]["y"],
        result["value"]["width"],
        result["value"]["height"],
    ]
    label_values = result["value"][result["type"]]
    return fol.Detection(
        label=label_values[0],
        bounding_box=_normalize_values(ls_box),
        rotation=result["value"].get("rotation", 0),
    )


def _from_polygonlabels(result):
    ls_points = _normalize_values(result["value"]["points"])
    label_values = result["value"][result["type"]]
    kwargs = dict(points=[ls_points], filled=True, closed=True)
    return fol.Polyline(label=label_values[0], **kwargs)


def _from_keypointlabels(result):
    keypoints = []
    group_fn = lambda x: x["value"]["keypointlabels"][0]
    for key, group in itertools.groupby(result, group_fn):
        points = [(one["value"]["x"], one["value"]["y"]) for one in group]
        points = _normalize_values(points)
        keypoints.append(fol.Keypoint(label=key, points=points))

    return keypoints


def _from_brushlabels(result):
    brush = fou.lazy_import(
        "label_studio_converter.brush",
        callback=lambda: fou.ensure_import("label_studio_converter.brush"),
    )

    label_values = result["value"]["brushlabels"]
    img = brush.decode_rle(result["value"]["rle"])
    mask = img.reshape(
        (result["original_height"], result["original_width"], 4)
    )[:, :, 3]
    return fol.Segmentation(label=label_values[0], mask=mask)


def _from_number(result):
    return fol.Regression(value=result["value"]["number"])


def _check_type(label, label_type, label_type_multiple):
    is_singular = isinstance(label, (label_type, label_type_multiple))
    is_list_type = isinstance(label, list) and isinstance(label[0], label_type)
    return is_singular or is_list_type


def _ls_tags_from_type(label_type):
    """Maps fiftyone types to Label Studio tags.

    Args:
        label_type: label studio type

    Returns:
        A tuple of parent tag, child tag and parent_tag kwargs
    """
    x = _LABEL_TYPES.get(
        label_type, _LABEL_TYPES[foua._RETURN_TYPES_MAP[label_type]]
    )
    return x["parent_tag"], x["child_tag"], x.get("tag_kwargs", {})


def _label_class_from_tag(label_type):
    """Maps Label Studio parent tag to fo.Label."""
    reverse = {
        v["parent_tag"].lower(): v["label"] for v in _LABEL_TYPES.values()
    }
    if label_type in reverse:
        return reverse[label_type]

    raise ValueError("Unknown label type: %s" % label_type)


def _tag_from_label(label_cls):
    label_cls = label_cls.lower()
    if label_cls in _LABEL_TO_TYPE:
        label_type = _LABEL_TO_TYPE[label_cls]
        return _LABEL_TYPES[label_type]["parent_tag"]

    if label_cls in foua._RETURN_TYPES_MAP:
        label_type = foua._RETURN_TYPES_MAP[label_cls]
        return _LABEL_TYPES[label_type]["parent_tag"]

    raise ValueError("Unsupported label class: %s" % label_cls)


def _normalize_values(values):
    values = np.array(values) / 100
    return values.tolist()


def _denormalize_values(values):
    values = np.array(values) * 100
    return values.tolist()


def _get_label_ids(label):
    if label is None:
        return []
    if isinstance(
        label, (fol.Classification, fol.Segmentation, fol.Regression)
    ):
        ids = label.id
    elif isinstance(label, fol.Classifications):
        ids = [l.id for l in label.classifications]
    elif isinstance(label, fol.Detections):
        ids = [l.id for l in label.detections]
    elif isinstance(label, fol.Polylines):
        ids = [l.id for l in label.polylines]
    elif isinstance(label, fol.Keypoints):
        ids = [l.id for l in label.keypoints]
    else:
        raise ValueError("Unsupported label type: %s" % type(label))

    return ids


_LABEL_TYPES = {
    "classification": dict(
        parent_tag="Choices",
        child_tag="Choice",
        label=fol.Classification,
        tag_kwargs={"choice": "single-radio"},
    ),
    "classifications": dict(
        parent_tag="Choices",
        child_tag="Choice",
        label=fol.Classification,
        tag_kwargs={"choice": "multiple"},
    ),
    "detections": dict(
        parent_tag="RectangleLabels",
        child_tag="Label",
        label=fol.Detection,
    ),
    "polylines": dict(
        parent_tag="PolygonLabels",
        child_tag="Label",
        label=fol.Polyline,
    ),
    "keypoints": dict(
        parent_tag="KeyPointLabels",
        child_tag="Label",
        label=fol.Keypoint,
    ),
    "segmentation": dict(
        parent_tag="BrushLabels",
        child_tag="Label",
        label=fol.Segmentation,
    ),
    "scalar": dict(
        parent_tag="Number",
        child_tag=None,
        label=fol.Regression,
    ),
}
_LABEL_TO_TYPE = {v["label"].__name__: k for k, v in _LABEL_TYPES.items()}



"""
Utilities for usage analytics.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import os
import multiprocessing
from socket import gaierror
import threading
import uuid

from httpx import HTTPError
import universal_analytics as ua

import fiftyone as fo
import fiftyone.constants as foc
from fiftyone.core.context import _get_context


_FIRST_IMPORT = "FIFTYONE_FIRST_IMPORT"
_import_logged = False


def get_user_id():
    """Gets the UUID of the current user.

    Returns:
        a tuple of

        -   UUID string
        -   True/False whether the UUID was newly created
    """
    uid_path = os.path.join(foc.FIFTYONE_CONFIG_DIR, "var", "uid")

    def read():
        try:
            with open(uid_path) as f:
                return next(f).strip()
        except (IOError, StopIteration):
            return None

    first_import = False
    if _FIRST_IMPORT in os.environ:
        first_import = os.environ[_FIRST_IMPORT] == "1"

    if not read():
        os.environ[_FIRST_IMPORT] = "1"
        os.makedirs(os.path.dirname(uid_path), exist_ok=True)
        with open(uid_path, "w") as f:
            f.write(str(uuid.uuid4()))

        return read(), True

    return read(), first_import


def log_import_if_allowed(test=False):
    """Logs the FiftyOne import event, if allowed.

    Args:
        test (False): whether to use the "test" uid
    """
    if fo.config.do_not_track:
        return

    if os.environ.get("FIFTYONE_DISABLE_SERVICES", None):
        return

    if os.environ.get("FIFTYONE_SERVER", None):
        return

    if multiprocessing.current_process().name != "MainProcess":
        return

    if test:
        uid, first_import = "test", False
    else:
        uid, first_import = get_user_id()

    kind = "new" if first_import else "returning"

    def send_import_event():
        try:
            with ua.HTTPRequest() as http:
                tracker = ua.Tracker(foc.UA_ID, http, client_id=uid)
                tracker.send(
                    "event",
                    "import",
                    kind,
                    label="%s-%s" % (foc.VERSION, _get_context()),
                )

            global _import_logged
            _import_logged = True
        except (gaierror, HTTPError) as e:
            pass

    th = threading.Thread(target=send_import_event)
    th.start()



"""
Utilities for working with the
`ActivityNet dataset <http://activity-net.org/index.html>`.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from itertools import chain
import logging
import os
import random

import eta.core.serial as etas
import eta.core.utils as etau
import eta.core.web as etaw

import fiftyone.core.utils as fou
import fiftyone.utils.data as foud
import fiftyone.utils.youtube as fouy


logger = logging.getLogger(__name__)


def download_activitynet_split(
    dataset_dir,
    split,
    source_dir=None,
    classes=None,
    max_duration=None,
    copy_files=True,
    num_workers=None,
    shuffle=None,
    seed=None,
    max_samples=None,
    version="200",
):
    """Utility that downloads full or partial splits of the
    `ActivityNet dataset <http://activity-net.org/index.html>`_.

    Args:
        dataset_dir: the directory to download the dataset
        split: the split to download. Supported values are
            ``("train", "validation", "test")``
        source_dir (None): the directory containing the manually downloaded
            ActivityNet files
        classes (None): a string or list of strings specifying required classes
            to load. If provided, only samples containing at least one instance
            of a specified class will be loaded
        max_duration (None): only videos with a duration in seconds that is
            less than or equal to the ``max_duration`` will be downloaded. By
            default, all videos are downloaded
        copy_files (True): whether to move (False) or create copies (True) of
            the source files when populating ``dataset_dir``. This is only
            relevant when a ``source_dir`` is provided
        num_workers (None): the number of threads to use when downloading
            individual video. By default, ``multiprocessing.cpu_count()`` is
            used
        shuffle (False): whether to randomly shuffle the order in which samples
            are chosen for partial downloads
        seed (None): a random seed to use when shuffling
        max_samples (None): a maximum number of samples to load per split. If
            ``classes`` are also specified, only up to the number of samples
            that contain at least one specified class will be loaded. By
            default, all matching samples are loaded
        version ("200"): the ActivityNet dataset version to download. The
            supported values are ``("100", "200")``

    Returns:
        a tuple of:

        -   **num_samples**: the total number of downloaded videos, or ``None``
            if everything was already downloaded
        -   **classes**: the list of all classes, or ``None`` if everything was
            already downloaded
        -   **did_download**: whether any content was downloaded (True) or if
            all necessary files were already downloaded (False)
    """
    manager = ActivityNetDatasetManager.from_dataset_dir(dataset_dir, version)

    if source_dir:
        manager.process_source(source_dir, copy_files)

    download_config = ActivityNetDownloadConfig(
        split,
        source_dir=source_dir,
        classes=classes,
        max_duration=max_duration,
        copy_files=copy_files,
        num_workers=num_workers,
        shuffle=shuffle,
        seed=seed,
        max_samples=max_samples,
    )

    all_classes = manager.all_classes
    num_existing = len(manager.existing_split_sample_ids(split))

    if download_config.load_entire_split:
        num_total = _NUM_TOTAL_SAMPLES[version][split]
        if num_existing != num_total:
            raise ValueError(
                "Found %d samples out of %d for split `%s`. In order to load "
                "a full ActivityNet split, you must download the source files "
                "from the ActivityNet maintainers. See "
                "https://voxel51.com/docs/fiftyone/integrations/activitynet.html#lactivitynet-full-split-downloads "
                "for more information."
                "\n\n"
                "Alternatively, provide the `max_samples`, `max_duration`, "
                "and/or `classes` parameters to download a subset of the "
                "dataset from YouTube." % (num_existing, num_total, split)
            )

        num_samples = num_existing
    else:
        (
            _,
            num_downloaded_samples,
        ) = manager.download_necessary_samples(download_config)
        num_samples = num_existing + num_downloaded_samples

    manager.write_data_json(split)

    return num_samples, all_classes


class ActivityNetDatasetImporter(
    foud.FiftyOneTemporalDetectionDatasetImporter
):
    """Class for importing AcitivityNet dataset splits downloaded via
    :meth:`download_activitynet_split`.

    Args:
        dataset_dir (None): the dataset directory
        data_path (None): an optional parameter that enables explicit control
            over the location of the media. Can be any of the following:

            -   a folder name like ``"data"`` or ``"data"/`` specifying a
                subfolder of ``dataset_dir`` where the media files reside
            -   an absolute directory path where the media files reside. In
                this case, the ``dataset_dir`` has no effect on the location of
                the data
            -   a filename like ``"data.json"`` specifying the filename of the
                JSON data manifest file in ``dataset_dir``
            -   an absolute filepath specifying the location of the JSON data
                manifest. In this case, ``dataset_dir`` has no effect on the
                location of the data

            If None, this parameter will default to whichever of ``data/`` or
            ``data.json`` exists in the dataset directory
        labels_path (None): an optional parameter that enables explicit control
            over the location of the labels. Can be any of the following:

            -   a filename like ``"labels.json"`` specifying the location of
                the labels in ``dataset_dir``
            -   an absolute filepath to the labels. In this case,
                ``dataset_dir`` has no effect on the location of the labels

            If None, the parameter will default to ``labels.json``
        classes (None): a string or list of strings specifying required classes
            to load. If provided, only samples containing at least one instance
            of a specified class will be loaded
        max_duration (None): only videos with a duration in seconds that is
            less than or equal to the `max_duration` will be loaded. By
            default, all videos are loaded
        compute_metadata (False): whether to produce
            :class:`fiftyone.core.metadata.VideoMetadata` instances for each
            video when importing
        shuffle (False): whether to randomly shuffle the order in which the
            samples are imported
        seed (None): a random seed to use when shuffling
        max_samples (None): a maximum number of samples to import. By default,
            all samples are imported
    """

    def __init__(
        self,
        dataset_dir=None,
        data_path=None,
        labels_path=None,
        classes=None,
        max_duration=None,
        compute_metadata=False,
        shuffle=False,
        seed=None,
        max_samples=None,
    ):
        super().__init__(
            dataset_dir=dataset_dir,
            data_path=data_path,
            labels_path=labels_path,
            compute_metadata=compute_metadata,
            shuffle=shuffle,
            seed=seed,
            max_samples=max_samples,
        )
        self.classes = classes
        self.max_duration = max_duration

        self.data_path = self._parse_data_path(
            dataset_dir=dataset_dir,
            data_path=data_path,
            default="data.json",
        )
        labels_path = self._parse_labels_path(
            dataset_dir=dataset_dir,
            labels_path=labels_path,
            default="labels.json",
        )
        self._taxonomy = None

    @property
    def has_dataset_info(self):
        return self._classes is not None or self._taxonomy is not None

    def setup(self):
        self._video_paths_map = self._load_data_map(
            self.data_path, ignore_exts=True, recursive=True
        )

        self._sample_parser = foud.FiftyOneTemporalDetectionSampleParser(
            compute_metadata=self.compute_metadata
        )

        process_uuids = True
        if self.labels_path is not None and os.path.isfile(self.labels_path):
            labels = etas.read_json(self.labels_path)
            info = ActivityNetInfo(labels)
            sample_ids = self._video_paths_map.keys()

            if self.classes or self.max_duration:
                # Load a subset of data
                any_sample_ids, all_sample_ids = info.get_matching_samples(
                    max_duration=self.max_duration,
                    classes=self.classes,
                    ids=[i[2:] for i in sample_ids],
                )
                any_sample_ids = list(any_sample_ids.keys())
                all_sample_ids = list(all_sample_ids.keys())

                process_uuids = False
                if not self.max_samples:
                    uuids = list(set(any_sample_ids + all_sample_ids))
                    self._uuids = self._preprocess_list(uuids)
                elif self.max_samples > len(all_sample_ids):
                    uuids = all_sample_ids
                    self.max_samples -= len(all_sample_ids)
                    any_sample_ids = self._preprocess_list(any_sample_ids)
                    uuids += any_sample_ids
                    self._uuids = uuids
                else:
                    self._uuids = self._preprocess_list(all_sample_ids)

                self._uuids = ["v_" + i for i in self._uuids]
                sample_ids = self._uuids

            labels = info.format_annotations([i[2:] for i in sample_ids])
        else:
            labels = {}

        self._taxonomy = labels.get("taxonomy", None)
        self._classes = labels.get("classes", None)
        self._labels_map = labels.get("labels", {})
        self._has_labels = any(self._labels_map.values())

        if process_uuids:
            uuids = sorted(self._labels_map.keys())
            self._uuids = self._preprocess_list(uuids)

        self._num_samples = len(self._uuids)

    def get_dataset_info(self):
        info = {}
        if self._classes is not None:
            info["classes"] = self._classes

        if self._taxonomy is not None:
            info["taxonomy"] = self._taxonomy

        return info


class ActivityNetDownloadConfig(object):
    """Configuration class for downloading full or partial splits from the
    ActivityNet dataset.
    """

    def __init__(
        self,
        split,
        source_dir=None,
        classes=None,
        max_duration=None,
        copy_files=True,
        num_workers=None,
        shuffle=None,
        seed=None,
        max_samples=None,
    ):
        self.split = split
        self.source_dir = source_dir
        self.classes = classes
        self.max_duration = max_duration
        self.copy_files = copy_files
        self.num_workers = num_workers
        self.shuffle = shuffle
        self.seed = seed
        self.max_samples = max_samples

        self.validate()

    @property
    def load_entire_split(self):
        if self.split != "test":
            classes = self.classes
        else:
            classes = None

        return bool(
            self.max_duration is None
            and self.max_samples is None
            and classes is None
        )

    def validate(self):
        self.validate_split()
        self.validate_max_duration()

    def validate_split(self):
        if self.split not in _SPLIT_MAP.keys():
            raise ValueError(
                "Unsupported split '%s'; supported values are %s"
                % (self.split, tuple(_SPLIT_MAP.keys()))
            )

    def validate_max_duration(self):
        if self.max_duration is not None and self.max_duration <= 0:
            raise ValueError(
                "`max_duration` must be a positive integer or float"
            )


class ActivityNetDatasetManager(object):
    """Class that manages the sample IDs and labels that need to be downloaded
    to load the specified subset of an ActivityNet dataset.
    """

    def __init__(self, foz_dir, version):
        self.version = version
        self.a100_info = ActivityNet100DatasetInfo(foz_dir)
        self.a200_info = ActivityNet200DatasetInfo(foz_dir)

    @property
    def info(self):
        if self.version == "100":
            return self.a100_info

        if self.version == "200":
            return self.a200_info

        raise ValueError(
            "Dataset version '%s' is not supported" % self.version
        )

    @property
    def all_classes(self):
        return self.info.all_classes

    def existing_split_sample_ids(self, split):
        return self.info.existing_split_sample_ids(split)

    def split_sample_ids(self, split):
        return self.info.split_sample_ids[split]

    def process_source(self, source_dir, copy_files):
        source_dir = os.path.expanduser(source_dir)
        for item in os.listdir(source_dir):
            if item in _SOURCE_ZIPS:
                logger.info("Processing source zip '%s'...", item)
                self._process_source_zip(item, source_dir, copy_files)
            elif item in _SOURCE_DIR_NAMES:
                logger.info("Processing source directory '%s'...", item)
                self._process_source_dir(item, source_dir, copy_files)

        self.a100_info.update_existing_sample_ids()
        self.a200_info.update_existing_sample_ids()

    def _process_source_zip(self, zip_name, source_dir, copy_files):
        zip_path = os.path.join(source_dir, zip_name)
        must_extract = self._check_zip_requirement(zip_name)

        if must_extract:
            zip_path = os.path.join(source_dir, zip_name)
            etau.extract_archive(zip_path, delete_archive=True)
            base_name = zip_name.replace(".zip", "").replace(".tar.gz", "")
            if "missing_files" not in base_name:
                base_name = "v1-2" if "v1-2" in base_name else "v1-3"

            self._process_source_dir(base_name, source_dir, copy_files)

    def _check_zip_requirement(self, zip_name):
        if zip_name == "missing_files.zip":
            version = "200"
            splits = ["train", "test", "validation"]
        else:
            version = "100" if "v1-2" in zip_name else "200"
            splits = []
            if "test" in zip_name:
                splits.append("test")

            if "train" in zip_name:
                splits.append("train")

            if "val" in zip_name:
                splits.append("validation")

        missing_splits = []
        for split in splits:
            missing_splits.append(
                self._split_is_missing_videos(version, split)
            )

        return any(missing_splits)

    def _split_is_missing_videos(self, version, split):
        if version == "100":
            info = self.a100_info
        else:
            info = self.a200_info

        existing_samples = info.existing_split_sample_ids(split)
        num_existing_samples = len(existing_samples)
        num_required_samples = _NUM_TOTAL_SAMPLES[version][split]
        return num_existing_samples != num_required_samples

    def _process_source_dir(self, dir_name, source_dir, copy_files):
        dir_list = []
        if dir_name == "missing_files":
            videos_dir = os.path.join(source_dir, dir_name)
            self._process_files(videos_dir, copy_files)
            dir_list.append((videos_dir, None, None))
        else:
            if "v1-2" in dir_name:
                version = "100"
            else:
                version = "200"

            if "missing_files" in dir_name:
                videos_dir = os.path.join(source_dir, dir_name)
                dir_list.append((videos_dir, version, "test"))
            else:
                version_dir = os.path.join(source_dir, dir_name)
                for split_dir in os.listdir(version_dir):
                    if split_dir == "train_val":
                        split = None
                    if split_dir == "val":
                        split = "validation"
                    else:
                        split = split_dir

                    videos_dir = os.path.join(version_dir, split_dir)
                    dir_list.append((videos_dir, version, split))

        for videos_dir, version, split in dir_list:
            self._process_files(
                videos_dir, copy_files, version=version, split=split
            )

    def _process_files(self, videos_dir, copy_files, version=None, split=None):
        existing_videos = self.a200_info.existing_sample_ids
        general_dest_dir = None
        if version is not None and split is not None:
            if version == "100":
                general_dest_dir = self.a100_info.data_dir(split)
            else:
                general_dest_dir = self.a200_info.data_dir(split)

        videos = os.listdir(videos_dir)
        if videos:
            with fou.ProgressBar() as pb:
                for video in pb(videos):
                    video_fn = os.path.splitext(video)[0]

                    # strip "v_" from video filename
                    video_id = video_fn[2:]
                    if video_id in existing_videos:
                        continue

                    if not general_dest_dir:
                        dest_dir = self._get_video_destination(
                            video_id, version=version, split=split
                        )
                    else:
                        dest_dir = general_dest_dir

                    if os.path.exists(os.path.join(dest_dir, video)):
                        continue

                    video_path = os.path.join(videos_dir, video)

                    if copy_files:
                        etau.copy_file(video_path, dest_dir)
                    else:
                        etau.move_file(video_path, dest_dir)

    def _get_video_destination(self, video_id, version=None, split=None):
        if version is None:
            version = self.a100_info.get_sample_dataset_version(video_id)

        if split is None:
            split = self.a200_info.get_sample_split(video_id)

        if version == "100":
            data_dir = self.a100_info.data_dir(split)
        else:
            data_dir = self.a200_info.data_dir(split)

        return data_dir

    def download_necessary_samples(self, config):
        prev_downloaded_ids = self.info.existing_sample_ids
        max_samples = config.max_samples
        seed = config.seed
        shuffle = config.shuffle
        num_workers = config.num_workers
        classes = config.classes
        max_duration = config.max_duration
        split = config.split

        self.info.cleanup_split(split)

        any_class_samples, all_class_samples = self.info.get_matching_samples(
            split=split,
            max_duration=max_duration,
            classes=classes,
        )

        all_class_ids = list(all_class_samples.keys())
        set_all_ids = set(all_class_ids)
        any_class_ids = list(any_class_samples.keys())
        set_any_ids = set(any_class_ids)

        requested_sample_ids = []
        requested_num = max_samples
        num_downloaded_samples = 0
        set_downloaded_ids = set(prev_downloaded_ids)

        if shuffle and seed is not None:
            random.seed(seed)

        # 1) Take the all class IDs that are downloaded up to max_samples
        dl_all_class_ids = list(set_all_ids.intersection(set_downloaded_ids))
        loaded_samples = self._load_requested_samples(
            dl_all_class_ids,
            shuffle,
            requested_num,
        )
        requested_sample_ids.extend(loaded_samples)
        num_loaded = len(loaded_samples)
        if requested_num is not None:
            requested_num -= num_loaded

        # 2) Take the any class IDs that are downloaded up to max_samples
        dl_any_class_ids = list(set_any_ids.intersection(set_downloaded_ids))
        loaded_samples = self._load_requested_samples(
            dl_any_class_ids,
            shuffle,
            requested_num,
        )
        requested_sample_ids.extend(loaded_samples)
        num_loaded = len(loaded_samples)
        if requested_num is not None:
            requested_num -= num_loaded

        # 3) Download all class IDs up to max_samples
        not_dl_all_class_ids = list(set_all_ids - set(dl_all_class_ids))
        downloaded_samples = self._download_requested_samples(
            not_dl_all_class_ids,
            all_class_samples,
            shuffle,
            requested_num,
            num_workers,
            split,
        )
        requested_sample_ids.extend(downloaded_samples)
        num_downloaded = len(downloaded_samples)
        num_downloaded_samples += num_downloaded
        if requested_num is not None:
            requested_num -= num_downloaded

        # 4) Download any class IDs up to max_samples
        not_dl_any_class_ids = list(set_any_ids - set(dl_any_class_ids))
        downloaded_samples = self._download_requested_samples(
            not_dl_any_class_ids,
            any_class_samples,
            shuffle,
            requested_num,
            num_workers,
            split,
        )
        requested_sample_ids.extend(downloaded_samples)
        num_downloaded = len(downloaded_samples)
        num_downloaded_samples += num_downloaded
        if requested_num is not None:
            requested_num -= num_downloaded

        self.a200_info.update_existing_sample_ids()

        return requested_sample_ids, num_downloaded_samples

    def _load_requested_samples(self, ids, shuffle, requested_num):
        if (requested_num is None or requested_num) and ids:
            ids = sorted(ids)
            if shuffle:
                random.shuffle(ids)

            if requested_num is not None:
                requested_sample_ids = ids[:requested_num]
            else:
                requested_sample_ids = ids

            return requested_sample_ids

        return []

    def _download_requested_samples(
        self, ids, class_samples, shuffle, requested_num, num_workers, split
    ):
        if (requested_num is None or requested_num) and ids:
            ids = sorted(ids)
            if shuffle:
                random.shuffle(ids)

            downloaded_ids = self._separate_versions_and_attempt_to_download(
                ids,
                class_samples,
                requested_num,
                num_workers,
                split,
            )

            return downloaded_ids

        return []

    def _separate_versions_and_attempt_to_download(
        self,
        ids,
        samples_info,
        num_samples,
        num_workers,
        split,
    ):
        all_a100_ids = self.a100_info.all_sample_ids
        a100_ids = [i for i in ids if i in all_a100_ids]
        a200_ids = list(set(ids) - set(a100_ids))

        downloaded_ids = []
        num_a100_ids = len(a100_ids)
        remaining_samples = num_samples

        if num_a100_ids:
            if num_samples is None:
                num_to_download = num_a100_ids
            else:
                num_to_download = min(num_a100_ids, num_samples)

            downloaded_ids, a100_errors = self._attempt_to_download(
                self.a100_info.data_dir(split),
                a100_ids,
                samples_info,
                num_to_download,
                num_workers,
            )
            if remaining_samples is not None:
                remaining_samples -= len(downloaded_ids)

            self._merge_and_write_errors(
                a100_errors, self.a100_info.error_path(split)
            )

        if remaining_samples:
            a200_downloaded_ids, a200_errors = self._attempt_to_download(
                self.a200_info.data_dir(split),
                a200_ids,
                samples_info,
                remaining_samples,
                num_workers,
            )
            downloaded_ids.extend(a200_downloaded_ids)
            self._merge_and_write_errors(
                a200_errors, self.a200_info.error_path(split)
            )

        return downloaded_ids

    def _attempt_to_download(
        self, videos_dir, ids, samples_info, num_samples, num_workers
    ):
        download_urls = []
        download_paths = []
        for sample_id in ids:
            sample_info = samples_info[sample_id]
            download_path = os.path.join(videos_dir, "v_%s.mp4" % sample_id)
            download_urls.append(sample_info["url"])
            download_paths.append(download_path)

        logger.info("Downloading videos from YouTube...")
        downloaded, errors = fouy.download_youtube_videos(
            urls=download_urls,
            video_paths=download_paths,
            max_videos=num_samples,
            num_workers=num_workers,
        )
        downloaded_ids = [ids[ind] for ind in downloaded.keys()]
        errors_dict = {}
        for ind, error in errors.items():
            errors_dict[download_urls[ind]] = error

        return downloaded_ids, errors_dict

    def _merge_and_write_errors(self, download_errors, error_path):
        if os.path.isfile(error_path):
            prev_errors = etas.read_json(error_path)
        else:
            prev_errors = {}

        for video, e in download_errors.items():
            if e in prev_errors:
                prev_errors[e].append(video)
            else:
                prev_errors[e] = [video]

        for e, videos in prev_errors.items():
            prev_errors[e] = sorted(set(videos))

        etas.write_json(prev_errors, error_path, pretty_print=True)

    def write_data_json(self, split):
        a100_start_dir = os.path.dirname(self.a100_info.data_json_path(split))
        a100_data_map = self._data_map(self.a100_info, split, a100_start_dir)
        if a100_data_map:
            etas.write_json(
                a100_data_map,
                self.a100_info.data_json_path(split),
                pretty_print=True,
            )

        start_dir = os.path.dirname(self.info.data_json_path(split))
        a100_data_map = self._data_map(self.a100_info, split, start_dir)
        a200_data_map = self._data_map(self.a200_info, split, start_dir)
        a200_data_map.update(a100_data_map)
        if a200_data_map:
            etas.write_json(
                a200_data_map,
                self.a200_info.data_json_path(split),
                pretty_print=True,
            )

    def _data_map(self, info, split, start_dir):
        to_rel = lambda fp, d: os.path.relpath(fp, start=d)
        to_uuid = lambda p: os.path.splitext(p)[0]
        data_path = info.data_dir(split)
        data_map = {
            to_uuid(p): to_rel(os.path.join(data_path, p), start_dir)
            for p in etau.list_files(data_path, recursive=True)
            if not (p.endswith(".part") or p.endswith(".ytdl"))
        }
        return data_map

    @classmethod
    def from_dataset_dir(cls, dataset_dir, version):
        dataset_dir = os.path.abspath(dataset_dir)
        foz_dir, _, _ = ActivityNetDatasetInfo.get_dir_info(dataset_dir)
        return cls(foz_dir, version)


class ActivityNetInfo(object):
    """Necessary information used to parse and format annotations."""

    def __init__(self, raw_annotations):
        self.raw_annotations = raw_annotations
        self.taxonomy = self.raw_annotations["taxonomy"]
        self.all_classes = _get_all_classes(self.taxonomy)

        self._splitwise_sample_ids = self._parse_sample_ids()
        self.all_sample_ids = _flatten_list(
            self._splitwise_sample_ids.values()
        )

    def _parse_sample_ids(self):
        ids = {}
        for anno_id, annot in self.raw_annotations["database"].items():
            split = _SPLIT_MAP_REV[annot["subset"]]
            if split not in ids:
                ids[split] = []
            ids[split].append(anno_id)
        return ids

    def get_matching_samples(
        self, split=None, max_duration=None, classes=None, ids=None
    ):
        if split:
            classes = self._parse_classes(classes, split)

        # Sample contains all specified classes
        all_class_match = {}

        # Sample contains any specified classes
        any_class_match = {}

        if split:
            activitynet_split = _SPLIT_MAP[split]
        else:
            activitynet_split = None

        if classes is not None:
            class_set = set(classes)

        for sample_id, anno_info in self.raw_annotations["database"].items():
            if ids and sample_id not in ids:
                continue

            if activitynet_split:
                is_correct_split = activitynet_split == anno_info["subset"]
            else:
                is_correct_split = True
            if max_duration:
                is_correct_dur = max_duration >= anno_info["duration"]
            else:
                is_correct_dur = True

            if not is_correct_split or not is_correct_dur:
                continue

            if classes is None:
                any_class_match[sample_id] = anno_info
            else:
                anno_labels = set(
                    [a["label"] for a in anno_info["annotations"]]
                )
                if class_set.issubset(anno_labels):
                    all_class_match[sample_id] = anno_info
                elif class_set & anno_labels:
                    any_class_match[sample_id] = anno_info

        return any_class_match, all_class_match

    def _parse_classes(self, classes, split):
        if classes is not None:
            if split == "test":
                logger.warning(
                    "Test split is unlabeled but `classes` were provided; "
                    "Skipping the split..."
                )

            bad_classes = list(set(classes) - set(self.all_classes))
            if bad_classes:
                raise ValueError(
                    "The following classes were specified but do not exist in "
                    "the dataset; %s" % bad_classes,
                )

        return classes

    def format_annotations(self, sample_ids, split=None):
        if split:
            activitynet_split = _SPLIT_MAP[split]
        else:
            activitynet_split = None

        labels = {}
        for anno_id, anno_info in self.raw_annotations["database"].items():
            if sample_ids is not None and anno_id not in sample_ids:
                continue

            if activitynet_split and anno_info["subset"] != activitynet_split:
                continue

            fo_anno_labels = []
            for an_anno_label in anno_info["annotations"]:
                target = an_anno_label["label"]
                timestamps = an_anno_label["segment"]
                fo_anno_labels.append(
                    {"label": target, "timestamps": timestamps}
                )

            sample_id = "v_" + anno_id
            labels[sample_id] = fo_anno_labels

        fo_annots = {
            "classes": self.all_classes,
            "labels": labels,
            "taxonomy": self.taxonomy,
        }
        return fo_annots


class ActivityNetSplitInfo(ActivityNetInfo):
    """Class that contains information related to paths, labels, and sample IDs
    of a single ActivityNet split.
    """

    def __init__(self, split_dir, version=None, raw_annotations=None):
        self.split_dir = os.path.abspath(split_dir)
        if raw_annotations:
            if not os.path.exists(self.raw_anno_path):
                etas.write_json(raw_annotations, self.raw_anno_path)
        else:
            raw_annotations = self._get_raw_annotations(version=version)

        self.raw_annotations = raw_annotations

        super().__init__(self.raw_annotations)

        self.update_existing_sample_ids()

    @property
    def raw_anno_path(self):
        return os.path.join(self.split_dir, "labels.json")

    @property
    def data_dir(self):
        _data_dir = os.path.join(self.split_dir, "data")
        etau.ensure_dir(_data_dir)
        return _data_dir

    @property
    def data_json_path(self):
        return os.path.join(self.split_dir, "data.json")

    @property
    def error_path(self):
        return os.path.join(self.split_dir, "download_errors.json")

    def update_existing_sample_ids(self):
        self.existing_sample_ids = self._get_video_files()

    def _get_video_files(self):
        video_filenames = etau.list_files(self.data_dir)
        video_ids = []
        for vfn in video_filenames:
            video_id, ext = os.path.splitext(vfn)
            if ext not in [".part", ".ytdl"]:
                vid = video_id[2:]
                video_ids.append(vid)

        return video_ids

    def cleanup(self):
        videos_dir = self.data_dir
        video_filenames = etau.list_files(videos_dir)
        for vfn in video_filenames:
            _, ext = os.path.splitext(vfn)
            if ext in [".part", ".ytdl"]:
                try:
                    os.remove(os.path.join(videos_dir, vfn))
                except FileNotFoundError:
                    pass

    def _get_raw_annotations(self, version=None):
        if not os.path.isfile(self.raw_anno_path):
            if not version:
                raise ValueError(
                    "Found `version=None` and no file at `%s`. If raw "
                    "annotations have not been loaded, then a version must be "
                    "provided." % self.raw_anno_path
                )

            anno_link = _ANNOTATION_DOWNLOAD_LINKS[version]
            etaw.download_file(anno_link, path=self.raw_anno_path)

        return etas.read_json(self.raw_anno_path)


class ActivityNetDatasetInfo(ActivityNetInfo):
    """Class that stores information related to paths, labels, and sample IDs
    for an ActivityNet dataset download.
    """

    def __init__(self, foz_dir):
        self.foz_dir = os.path.abspath(foz_dir)
        self._split_infos = {}
        self.raw_annotations = self._get_raw_annotations()
        super().__init__(self.raw_annotations)

        self.update_existing_sample_ids()

    def split_info(self, split):
        if split not in self._split_infos:
            self._split_infos[split] = ActivityNetSplitInfo(
                self.split_dir(split),
                version=self.version,
                raw_annotations=self.raw_annotations,
            )

        return self._split_infos[split]

    @property
    def splits(self):
        return ["train", "test", "validation"]

    @property
    def version(self):
        raise NotImplementedError("Subclass must implement version")

    @property
    def dataset_dir(self):
        return os.path.join(self.foz_dir, "activitynet-%s" % self.version)

    @property
    def raw_anno_path(self):
        return os.path.join(self.dataset_dir, "labels.json")

    def split_sample_ids(self, split):
        return self._splitwise_sample_ids[split]

    def existing_split_sample_ids(self, split):
        return self._splitwise_existing_sample_ids[split]

    def split_dir(self, split):
        return os.path.join(self.dataset_dir, split)

    def data_dir(self, split):
        return self.split_info(split).data_dir

    def data_json_path(self, split):
        return self.split_info(split).data_json_path

    def error_path(self, split):
        return self.split_info(split).error_path

    def update_existing_sample_ids(self):
        self._splitwise_existing_sample_ids = None
        self.existing_sample_ids = None
        raise NotImplementedError(
            "Subclass must implement `update_existing_sample_ids()`"
        )

    def _get_existing_sample_ids(self):
        ids = {}
        for split in self.splits:
            if os.path.exists(self.split_dir(split)):
                split_info = self.split_info(split)
                split_info.update_existing_sample_ids()
                split_ids = split_info.existing_sample_ids
            else:
                split_ids = []

            ids[split] = split_ids

        return ids

    def cleanup_split(self, split):
        self.split_info(split).cleanup()

    def _get_raw_annotations(self):
        if not os.path.isfile(self.raw_anno_path):
            anno_link = _ANNOTATION_DOWNLOAD_LINKS[self.version]
            etaw.download_file(anno_link, path=self.raw_anno_path)

        return etas.read_json(self.raw_anno_path)

    @classmethod
    def get_dir_info(cls, dataset_dir):
        if not os.path.basename(dataset_dir):
            dataset_dir = os.path.dirname(dataset_dir)

        split = os.path.basename(dataset_dir)
        an_dir = os.path.dirname(dataset_dir)
        an_dirname = os.path.basename(an_dir)
        an_version = cls._get_version_from_dirname(an_dirname)
        foz_dir = os.path.dirname(an_dir)
        return foz_dir, split, an_version

    @classmethod
    def _get_version_from_dirname(cls, an_dirname):
        version = an_dirname.split("-")[-1]
        return version

    def get_sample_split(self, sample_id):
        database = self.raw_annotations["database"]
        if sample_id not in database:
            return None

        an_split = database[sample_id]["subset"]
        return _SPLIT_MAP_REV[an_split]

    def get_sample_dataset_version(self, sample_id):
        raise NotImplementedError(
            "Subclass must implement `get_sample_dataset_version()`"
        )


class ActivityNet100DatasetInfo(ActivityNetDatasetInfo):
    """ActivityNet 100 dataset info."""

    @property
    def version(self):
        return "100"

    def get_sample_dataset_version(self, sample_id):
        if sample_id in self.all_sample_ids:
            return "100"

        return "200"

    def update_existing_sample_ids(self):
        self._splitwise_existing_sample_ids = self._get_existing_sample_ids()
        self.existing_sample_ids = _flatten_list(
            self._splitwise_existing_sample_ids.values()
        )


class ActivityNet200DatasetInfo(ActivityNetDatasetInfo):
    """ActivityNet 200 dataset info."""

    def __init__(self, foz_dir):
        self.a100_info = ActivityNet100DatasetInfo(foz_dir)
        super().__init__(foz_dir)

    @property
    def version(self):
        return "200"

    def get_sample_dataset_version(self, sample_id):
        if sample_id in self.a100_info.all_sample_ids:
            return "100"

        return "200"

    def update_existing_sample_ids(self):
        self._splitwise_existing_sample_ids = self._get_existing_sample_ids()
        self.a100_info.update_existing_sample_ids()

        for split in self.splits:
            self._splitwise_existing_sample_ids[split].extend(
                self.a100_info.existing_split_sample_ids(split)
            )

        self.existing_sample_ids = _flatten_list(
            self._splitwise_existing_sample_ids.values()
        )


def _get_all_classes(taxonomy):
    classes = set()
    parents = set()
    for node in taxonomy:
        node_name = node["nodeName"]
        parent_name = node["parentName"]
        classes.add(node_name)
        parents.add(parent_name)

    classes = sorted(classes - parents)

    return classes


def _flatten_list(l):
    l = [list(i) for i in l]
    return list(chain(*l))


_ANNOTATION_DOWNLOAD_LINKS = {
    "100": "http://ec2-52-25-205-214.us-west-2.compute.amazonaws.com/files/activity_net.v1-2.min.json",
    "200": "http://ec2-52-25-205-214.us-west-2.compute.amazonaws.com/files/activity_net.v1-3.min.json",
}

_SPLIT_MAP = {
    "train": "training",
    "test": "testing",
    "validation": "validation",
}

_SPLIT_MAP_REV = {v: k for k, v in _SPLIT_MAP.items()}

_NUM_TOTAL_SAMPLES = {
    "100": {
        "train": 4819,
        "test": 2480,
        "validation": 2383,
    },
    "200": {
        "train": 10024,
        "test": 5044,
        "validation": 4926,
    },
}

_SOURCE_ZIPS = [
    "missing_files.zip",
    "missing_files_v1-2_test.zip",
    "missing_files_v1-3_test.zip",
    "v1-2_test.tar.gz",
    "v1-2_train.tar.gz",
    "v1-2_val.tar.gz",
    "v1-3_test.tar.gz",
    "v1-3_train_val.tar.gz",
]

_SOURCE_DIR_NAMES = {
    "missing_files": [],
    "missing_files_v1-2_test": [],
    "missing_files_v1-3_test": [],
    "v1-2": ["test", "train", "val"],
    "v1-3": ["test", "train_val"],
}



"""
Base classes for objects that are backed by database documents.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from copy import deepcopy

from bson import ObjectId

import eta.core.serial as etas
import eta.core.utils as etau

import fiftyone.core.labels as fol
from fiftyone.core.singletons import DocumentSingleton


class _Document(object):
    def __init__(self, doc, dataset=None):
        self._doc = doc
        self._dataset = dataset

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return self._doc.fancy_repr(class_name=self.__class__.__name__)

    def __dir__(self):
        return super().__dir__() + list(self.field_names)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return self._doc == other._doc

    def __contains__(self, name):
        return self.has_field(name)

    def __getattr__(self, name):
        return self.get_field(name)

    def __setattr__(self, name, value):
        if name.startswith("_"):
            super().__setattr__(name, value)
            return

        if not self.has_field(name):
            dtype = "frame" if "Frame" in self.__class__.__name__ else "sample"
            raise ValueError(
                "Adding %s fields using the `%s.field = value` syntax is not "
                "allowed; use `%s['field'] = value` instead"
                % (dtype, dtype, dtype)
            )

        self.set_field(name, value)

    def __delattr__(self, name):
        try:
            super().__delattr__(name)
        except AttributeError:
            self.clear_field(name)

    def __getitem__(self, field_name):
        try:
            return self.get_field(field_name)
        except AttributeError as e:
            raise KeyError(e.args[0])

    def __setitem__(self, field_name, value):
        self.set_field(field_name, value)

    def __delitem__(self, field_name):
        try:
            self.clear_field(field_name)
        except AttributeError as e:
            raise KeyError(e.args[0])

    def __copy__(self):
        return self.copy()

    @property
    def _id(self):
        """The ObjectId of the document, or ``None`` if it has not been added
        to the database.
        """
        _id = self._doc.id
        return ObjectId(_id) if _id is not None else None

    @property
    def in_dataset(self):
        """Whether the document has been added to a dataset."""
        return self.dataset is not None

    @property
    def dataset(self):
        """The dataset to which this document belongs, or ``None`` if it has
        not been added to a dataset.
        """
        return self._dataset

    @property
    def _collection(self):
        """The :class:`fiftyone.core.collections.SampleCollection` from which
        this document was taken, or ``None`` if it is not in a dataset.
        """
        return self._dataset

    @property
    def field_names(self):
        """An ordered tuple of the public field names of this document."""
        return self._doc.field_names

    @property
    def _in_db(self):
        """Whether the document has been inserted into the database."""
        return self._doc.in_db

    def _get_field_names(self, include_private=False, use_db_fields=False):
        """Returns an ordered tuple of field names of this document.

        Args:
            include_private (False): whether to include private fields
            use_db_fields (False): whether to return database fields

        Returns:
            a tuple of field names
        """
        return self._doc._get_field_names(
            include_private=include_private,
            use_db_fields=use_db_fields,
        )

    def has_field(self, field_name):
        """Determines whether the document has the given field.

        Args:
            field_name: the field name

        Returns:
            True/False
        """
        return self._doc.has_field(field_name)

    def get_field(self, field_name):
        """Gets the value of a field of the document.

        Args:
            field_name: the field name

        Returns:
            the field value

        Raises:
            AttributeError: if the field does not exist
        """
        try:
            value = self._doc.get_field(field_name)
        except AttributeError:
            raise AttributeError(
                "%s has no field '%s'" % (self.__class__.__name__, field_name)
            )

        if isinstance(value, ObjectId):
            value = str(value)

        return value

    def set_field(
        self,
        field_name,
        value,
        create=True,
        validate=True,
        dynamic=False,
    ):
        """Sets the value of a field of the document.

        Args:
            field_name: the field name
            value: the field value
            create (True): whether to create the field if it does not exist
            validate (True): whether to validate values for existing fields
            dynamic (False): whether to declare dynamic embedded document
                fields

        Raises:
            ValueError: if ``field_name`` is not an allowed field name
            AttirubteError: if the field does not exist and ``create == False``
        """
        self._doc.set_field(
            field_name,
            value,
            create=create,
            validate=validate,
            dynamic=dynamic,
        )

    def update_fields(
        self,
        fields_dict,
        expand_schema=True,
        validate=True,
        dynamic=False,
    ):
        """Sets the dictionary of fields on the document.

        Args:
            fields_dict: a dict mapping field names to values
            expand_schema (True): whether to dynamically add new fields
                encountered to the document schema. If False, an error is
                raised if any fields are not in the document schema
            validate (True): whether to validate values for existing fields
            dynamic (False): whether to declare dynamic embedded document
                fields

        Raises:
            AttributeError: if ``expand_schema == False`` and a field does not
                exist
        """
        for field_name, value in fields_dict.items():
            self.set_field(
                field_name,
                value,
                create=expand_schema,
                validate=validate,
                dynamic=dynamic,
            )

    def clear_field(self, field_name):
        """Clears the value of a field of the document.

        Args:
            field_name: the name of the field to clear

        Raises:
            AttributeError: if the field does not exist
        """
        self._doc.clear_field(field_name)

    def iter_fields(self, include_id=False):
        """Returns an iterator over the ``(name, value)`` pairs of the public
        fields of the document.

        Args:
            include_id (False): whether to include the ``id`` field

        Returns:
            an iterator that emits ``(name, value)`` tuples
        """
        for field_name in self.field_names:
            if field_name == "id" and not include_id:
                continue

            yield field_name, self.get_field(field_name)

    def merge(
        self,
        document,
        fields=None,
        omit_fields=None,
        merge_lists=True,
        overwrite=True,
        expand_schema=True,
        validate=True,
        dynamic=False,
    ):
        """Merges the fields of the document into this document.

        The behavior of this method is highly customizable. By default, all
        top-level fields from the provided document are merged in, overwriting
        any existing values for those fields, with the exception of list fields
        (e.g., ``tags``) and label list fields (e.g.,
        :class:`fiftyone.core.labels.Detections` fields), in which case the
        elements of the lists themselves are merged. In the case of label list
        fields, labels with the same ``id`` in both documents are updated
        rather than duplicated.

        To avoid confusion between missing fields and fields whose value is
        ``None``, ``None``-valued fields are always treated as missing while
        merging.

        This method can be configured in numerous ways, including:

        -   Whether new fields can be added to the document schema
        -   Whether list fields should be treated as ordinary fields and merged
            as a whole rather than merging their elements
        -   Whether to merge only specific fields, or all but certain fields
        -   Mapping input document fields to different field names of this
            document

        Args:
            document: a :class:`Document` or :class:`DocumentView` of the same
                type
            fields (None): an optional field or iterable of fields to which to
                restrict the merge. This can also be a dict mapping field names
                of the input document to field names of this document
            omit_fields (None): an optional field or iterable of fields to
                exclude from the merge
            merge_lists (True): whether to merge the elements of top-level list
                fields (e.g., ``tags``) and label list fields (e.g.,
                :class:`fiftyone.core.labels.Detections` fields) rather than
                merging the entire top-level field like other field types.
                For label lists fields, existing
                :class:`fiftyone.core.label.Label` elements are either replaced
                (when ``overwrite`` is True) or kept (when ``overwrite`` is
                False) when their ``id`` matches a label from the provided
                document
            overwrite (True): whether to overwrite (True) or skip (False)
                existing fields and label elements
            expand_schema (True): whether to dynamically add new fields
                encountered to the document schema. If False, an error is
                raised if any fields are not in the document schema
            validate (True): whether to validate values for existing fields
            dynamic (False): whether to declare dynamic embedded document
                fields

        Raises:
            AttributeError: if ``expand_schema == False`` and a field does not
                exist
        """
        if not overwrite:
            existing_field_names = set(self.field_names)

        fields = document._parse_fields(fields=fields, omit_fields=omit_fields)

        for src_field, dst_field in fields.items():
            value = document[src_field]

            if value is None:
                continue

            try:
                curr_value = self[dst_field]
            except KeyError:
                curr_value = None

            if merge_lists:
                field_type = type(curr_value)

                if issubclass(field_type, list):
                    _merge_lists(curr_value, value, overwrite=overwrite)
                    continue

                if field_type in fol._LABEL_LIST_FIELDS:
                    _merge_labels(curr_value, value, overwrite=overwrite)
                    continue

            if (
                not overwrite
                and dst_field in existing_field_names
                and curr_value is not None
            ):
                continue

            self.set_field(
                dst_field,
                value,
                create=expand_schema,
                validate=validate,
                dynamic=dynamic,
            )

    def copy(self, fields=None, omit_fields=None):
        """Returns a deep copy of the document that has not been added to the
        database.

        Args:
            fields (None): an optional field or iterable of fields to which to
                restrict the copy. This can also be a dict mapping existing
                field names to new field names
            omit_fields (None): an optional field or iterable of fields to
                exclude from the copy

        Returns:
            a :class:`Document`
        """
        raise NotImplementedError("subclass must implement copy()")

    def to_dict(self, include_private=False):
        """Serializes the document to a JSON dictionary.

        Args:
            include_private (False): whether to include private fields

        Returns:
            a JSON dict
        """
        d = self._doc.to_dict(extended=True)

        if include_private:
            return d

        return {k: v for k, v in d.items() if not k.startswith("_")}

    def to_mongo_dict(self, include_id=False):
        """Serializes the document to a BSON dictionary equivalent to the
        representation that would be stored in the database.

        Args:
            include_id (False): whether to include the document ID

        Returns:
            a BSON dict
        """
        d = self._doc.to_dict()
        if not include_id:
            d.pop("_id", None)

        return d

    def to_json(self, pretty_print=False):
        """Serializes the document to a JSON string.

        The document ID and private fields are excluded in this representation.

        Args:
            pretty_print (False): whether to render the JSON in human readable
                format with newlines and indentations

        Returns:
            a JSON string
        """
        return etas.json_to_str(self.to_dict(), pretty_print=pretty_print)

    def save(self):
        """Saves the document to the database."""
        self._save()

    def _save(self, deferred=False):
        if not self._in_db:
            raise ValueError(
                "Cannot save a document that has not been added to a dataset"
            )

        return self._doc._save(deferred=deferred)

    def _parse_fields(self, fields=None, omit_fields=None):
        if fields is None:
            fields = {f: f for f in self.field_names if f != "id"}
        elif etau.is_str(fields):
            fields = {fields: fields}

        if not isinstance(fields, dict):
            fields = {f: f for f in fields}

        if omit_fields is not None:
            if etau.is_str(omit_fields):
                omit_fields = {omit_fields}
            else:
                omit_fields = set(omit_fields)

            fields = {k: v for k, v in fields.items() if k not in omit_fields}

        return fields


class Document(_Document):
    """Abstract base class for objects that are associated with
    :class:`fiftyone.core.dataset.Dataset` instances and are backed by
    documents in database collections.

    Document subclasses whose in-dataset instances should be singletons can
    inherit this behavior by deriving from the
    :class:`fiftyone.core.singletons.DocumentSingleton` metaclass.

    Args:
        **kwargs: field names and values
    """

    # The :class:`fiftyone.core.odm.document.Document` class used by this class
    # to store backing documents for instances that are *not* in the dataset
    _NO_DATASET_DOC_CLS = None

    def __init__(self, **kwargs):
        # pylint: disable=not-callable
        doc = self._NO_DATASET_DOC_CLS(**kwargs)
        super().__init__(doc)

    def copy(self, fields=None, omit_fields=None):
        fields = self._parse_fields(fields=fields, omit_fields=omit_fields)
        return self.__class__(
            **{v: deepcopy(self[k]) for k, v in fields.items()}
        )

    def reload(self, hard=False):
        """Reloads the document from the database.

        Args:
            hard (False): whether to reload the document's schema in addition
                to its field values. This is necessary if new fields may have
                been added to the document schema
        """
        if hard:
            self._reload_backing_doc()
        else:
            # We can only reload fields that are in our schema
            self._doc.reload(*list(self._doc))

    @classmethod
    def from_doc(cls, doc, dataset=None):
        """Creates a document backed by the given database document.

        Args:
            doc: a :class:`fiftyone.core.odm.document.Document`
            dataset (None): the :class:`fiftyone.core.dataset.Dataset` that
                the document belongs to, if any

        Returns:
            a :class:`Document`
        """
        if isinstance(doc, cls._NO_DATASET_DOC_CLS):
            document = cls.__new__(cls)
            document._doc = doc
            document._dataset = None
            return document

        if issubclass(type(cls), DocumentSingleton):
            document = cls._get_instance(doc)
            if document is not None:
                return document

        if dataset is None:
            raise ValueError(
                "`dataset` argument must be provided for documents in "
                "datasets"
            )

        document = cls.__new__(cls)
        document._doc = None  # prevents recursion
        document._set_backing_doc(doc, dataset=dataset)

        return document

    @classmethod
    def from_dict(cls, d):
        """Loads the document from a JSON dictionary.

        The returned document will not belong to a dataset.

        Returns:
            a :class:`Document`
        """
        doc = cls._NO_DATASET_DOC_CLS.from_dict(d, extended=True)
        return cls.from_doc(doc)

    @classmethod
    def from_json(cls, s):
        """Loads the document from a JSON string.

        The returned document will not belong to a dataset.

        Args:
            s: the JSON string

        Returns:
            a :class:`Document`
        """
        doc = cls._NO_DATASET_DOC_CLS.from_json(s)
        return cls.from_doc(doc)

    def _reload_backing_doc(self):
        """Reloads the backing doc from the database."""
        raise NotImplementedError(
            "subclass must implement _reload_backing_doc()"
        )

    def _set_backing_doc(self, doc, dataset=None):
        """Sets the backing doc for the document.

        Args:
            doc: a :class:`fiftyone.core.odm.document.Document`
            dataset (None): the :class:`fiftyone.core.dataset.Dataset` to which
                the document belongs, if any
        """
        self._doc = doc
        self._dataset = dataset

        cls = self.__class__
        if issubclass(type(cls), DocumentSingleton):
            cls._register_instance(self)

    def _reset_backing_doc(self):
        """Resets the backing doc for the document.

        The document will no longer belong to a dataset.
        """
        self._doc = self.copy()._doc
        self._dataset = None


class DocumentView(_Document):
    """A view into a :class:`Document` in a dataset.

    Like :class:`Document` instances, the fields of a :class:`DocumentView`
    instance can be modified, new fields can be created, and any changes can be
    saved to the database.

    :class:`DocumentView` instances differ from :class:`Document` instances
    in the following ways:

    -   A document view may contain only a subset of the fields of its source
        document, either by selecting and/or excluding specific fields
    -   A document view may contain array fields or embedded array fields that
        have been filtered, thus containing only a subset of the array elements
        from the source document
    -   Excluded fields of a document view may not be accessed or modified

    .. note::

        :meth:`DocumentView.save` will not delete any excluded fields or
        filtered array elements from the source document.

    Document views should never be created manually; they are generated when
    accessing the contents of a :class:`fiftyone.core.view.DatasetView`.

    Args:
        doc: a :class:`fiftyone.core.odm.document.Document`
        view: the :class:`fiftyone.core.view.DatasetView` that the document
            belongs to
        selected_fields (None): a set of field names that this document view is
            restricted to, if any
        excluded_fields (None): a set of field names that are excluded from
            this document view, if any
        filtered_fields (None): a set of field names of array fields that are
            filtered in this document view, if any
    """

    # The `Document` class associated with this `DocumentView` class
    # Subclasses must define this
    _DOCUMENT_CLS = Document

    def __init__(
        self,
        doc,
        view,
        selected_fields=None,
        excluded_fields=None,
        filtered_fields=None,
    ):
        if selected_fields is not None and excluded_fields is not None:
            selected_fields = selected_fields.difference(excluded_fields)
            excluded_fields = None

        self._view = view
        self._selected_fields = selected_fields
        self._excluded_fields = excluded_fields
        self._filtered_fields = filtered_fields

        super().__init__(doc, dataset=view._dataset)

    def __repr__(self):
        return self._doc.fancy_repr(
            class_name=self.__class__.__name__,
            select_fields=self._selected_fields,
            exclude_fields=self._excluded_fields,
        )

    @property
    def _collection(self):
        return self._view

    @property
    def field_names(self):
        """An ordered tuple of field names of this document view.

        This may be a subset of all fields of the document if fields have been
        selected or excluded.
        """
        return self._get_field_names(include_private=False)

    def _get_field_names(self, include_private=False, use_db_fields=False):
        field_names = super()._get_field_names(include_private=include_private)

        if self._selected_fields is not None:
            field_names = tuple(
                fn for fn in field_names if fn in self._selected_fields
            )

        if self._excluded_fields is not None:
            field_names = tuple(
                fn for fn in field_names if fn not in self._excluded_fields
            )

        if use_db_fields:
            return self._to_db_fields(field_names)

        return field_names

    def _to_db_fields(self, field_names):
        return self._doc._to_db_fields(field_names)

    @property
    def selected_field_names(self):
        """The set of field names that are selected on this document view, or
        ``None`` if no fields are explicitly selected.
        """
        return self._selected_fields

    @property
    def excluded_field_names(self):
        """The set of field names that are excluded on this document view, or
        ``None`` if no fields are explicitly excluded.
        """
        return self._excluded_fields

    @property
    def filtered_field_names(self):
        """The set of field names or ``embedded.field.names`` that have been
        filtered on this document view, or ``None`` if no fields are filtered.
        """
        return self._filtered_fields

    def has_field(self, field_name):
        ef = self._excluded_fields
        if ef is not None and field_name in ef:
            return False

        sf = self._selected_fields
        if sf is not None and field_name not in sf:
            return False

        return super().has_field(field_name)

    def get_field(self, field_name):
        ef = self._excluded_fields
        if ef is not None and field_name in ef:
            raise AttributeError(
                "Field '%s' is excluded from this %s"
                % (field_name, self.__class__.__name__)
            )

        value = super().get_field(field_name)

        sf = self._selected_fields
        if sf is not None and field_name not in sf:
            raise AttributeError(
                "Field '%s' was not selected on this %s"
                % (field_name, self.__class__.__name__)
            )

        return value

    def set_field(
        self,
        field_name,
        value,
        create=True,
        validate=True,
        dynamic=False,
    ):
        if not create:
            # Ensures field exists
            _ = self.get_field(field_name)

            super().set_field(
                field_name,
                value,
                create=create,
                validate=validate,
                dynamic=dynamic,
            )
        else:
            super().set_field(
                field_name,
                value,
                create=create,
                validate=validate,
                dynamic=dynamic,
            )

            if self._excluded_fields is not None:
                self._excluded_fields.discard(field_name)

            if self._selected_fields is not None:
                self._selected_fields.add(field_name)

    def clear_field(self, field_name):
        # Ensures field exists
        _ = self.get_field(field_name)

        super().clear_field(field_name)

    def to_dict(self, include_private=False):
        d = super().to_dict(include_private=include_private)

        if self._selected_fields or self._excluded_fields:
            field_names = set(
                self._get_field_names(
                    include_private=include_private,
                    use_db_fields=True,
                )
            )

            d = {k: v for k, v in d.items() if k in field_names}

        return d

    def to_mongo_dict(self, include_id=False):
        d = super().to_mongo_dict(include_id=include_id)

        if self._selected_fields or self._excluded_fields:
            field_names = set(
                self._get_field_names(include_private=True, use_db_fields=True)
            )

            d = {k: v for k, v in d.items() if k in field_names}

        return d

    def copy(self, fields=None, omit_fields=None):
        fields = self._parse_fields(fields=fields, omit_fields=omit_fields)
        return self._DOCUMENT_CLS(
            **{v: deepcopy(self[k]) for k, v in fields.items()}
        )

    def save(self):
        """Saves the document view to the database."""
        self._save()
        self._reload_parents()

    def _save(self, deferred=False):
        return self._doc._save(
            deferred=deferred,
            filtered_fields=self._filtered_fields,
        )

    def _reload_parents(self):
        if issubclass(type(self._DOCUMENT_CLS), DocumentSingleton):
            self._DOCUMENT_CLS._reload_instance(self)


def _merge_lists(dst, src, overwrite=True):
    if src is None:
        return

    dst.extend(v for v in src if v not in dst)


def _merge_labels(dst, src, overwrite=True):
    if src is None:
        return

    label_type = type(dst)
    list_field = label_type._LABEL_LIST_FIELD

    labels = dst[list_field]
    new_labels = src[list_field]

    if overwrite:
        existing_ids = {l.id: idx for idx, l in enumerate(labels)}
        for l in new_labels:
            idx = existing_ids.get(l.id, None)
            if idx is not None:
                labels[idx] = l
            else:
                labels.append(l)
    else:
        existing_ids = set(l.id for l in labels)
        labels.extend(l for l in new_labels if l.id not in existing_ids)



"""
Utilities for working with
`Lightning Flash <https://github.com/PyTorchLightning/lightning-flash>`_.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import inspect
import itertools

import numpy as np

import fiftyone.core.labels as fol
import fiftyone.core.media as fom
import fiftyone.core.utils as fou

fou.ensure_import("flash>=0.7dev")
import flash
import flash.core.classification as fc
import flash.image as fi
import flash.image.detection.output as fdo
import flash.image.segmentation.output as fso
import flash.video as fv


_SUPPORTED_MODELS = (
    fi.ImageClassifier,
    fi.ObjectDetector,
    fi.SemanticSegmentation,
)

_SUPPORTED_EMBEDDERS = fi.ImageEmbedder


def apply_flash_model(
    samples,
    model,
    label_field="predictions",
    confidence_thresh=None,
    store_logits=False,
    batch_size=None,
    num_workers=None,
    transform_kwargs=None,
    trainer_kwargs=None,
):
    """Applies the given
    :class:`Lightning Flash model <flash:flash.core.model.Task>` to the samples
    in the collection.

    Args:
        samples: a :class:`fiftyone.core.collections.SampleCollection`
        model: a :class:`flash:flash.core.model.Task`
        label_field ("predictions"): the name of the field in which to store
            the model predictions. When performing inference on video frames,
            the "frames." prefix is optional
        confidence_thresh (None): an optional confidence threshold to apply to
            any applicable labels generated by the model
        store_logits (False): whether to store logits for the model
            predictions. This is only supported when the provided ``model`` has
            logits
        batch_size (None): an optional batch size to use. If not provided, a
            default batch size is used
        num_workers (None): the number of workers for the data loader to use
        transform_kwargs (None): an optional dict of transform kwargs to pass
            into the created data module used by some models
        trainer_kwargs (None): an optional dict of kwargs used to initialize the
            :mod:`Trainer <flash:flash.core.trainer>`. These can be used to,
            for example, configure the number of GPUs to use and other
            distributed inference parameters
    """
    output = _get_output(model, confidence_thresh, store_logits)

    if trainer_kwargs is None:
        trainer_kwargs = {}

    data_kwargs = {
        "num_workers": num_workers or 1,
        "batch_size": batch_size or 1,
    }
    if transform_kwargs:
        data_kwargs["transform_kwargs"] = transform_kwargs

    datamodule_cls = _MODEL_TO_DATAMODULE_MAP[type(model)]

    datamodule = datamodule_cls.from_fiftyone(
        predict_dataset=samples, **data_kwargs
    )
    predictions = flash.Trainer(**trainer_kwargs).predict(
        model, datamodule=datamodule, output=output
    )
    predictions = list(itertools.chain.from_iterable(predictions))
    predictions = {p["filepath"]: p["predictions"] for p in predictions}

    samples.set_values(label_field, predictions, key_field="filepath")


def compute_flash_embeddings(
    samples,
    model,
    embeddings_field=None,
    batch_size=None,
    num_workers=None,
    transform_kwargs=None,
    trainer_kwargs=None,
):
    """Computes embeddings for the samples in the collection using the given
    :class:`Lightning Flash model <flash:flash.core.model.Task>`.

    This method only supports applying an
    :ref:`ImageEmbedder <flash:image_embedder>` to an image collection.

    If an ``embeddings_field`` is provided, the embeddings are saved to the
    samples; otherwise, the embeddings are returned in-memory.

    Args:
        samples: a :class:`fiftyone.core.collections.SampleCollection`
        model: a :class:`flash:flash.core.model.Task`
        embeddings_field (None): the name of a field in which to store the
            embeddings
        batch_size (None): an optional batch size to use. If not provided, a
            default batch size is used
        num_workers (None): the number of workers for the data loader to use
        transform_kwargs (None): an optional dict of transform kwargs to pass
            into the created data module used by some models
        trainer_kwargs (None): an optional dict of kwargs used to initialize the
            :mod:`Trainer <flash:flash.core.trainer>`. These can be used to,
            for example, configure the number of GPUs to use and other
            distributed inference parameters

    Returns:
        one of the following:

        -   ``None``, if an ``embeddings_field`` is provided
        -   a ``num_samples x num_dim`` array of embeddings, if
            ``embeddings_field`` is None
    """
    if not isinstance(model, _SUPPORTED_EMBEDDERS):
        raise ValueError(
            "Unsupported model type %s. Supported model types are %s"
            % (type(model), _SUPPORTED_EMBEDDERS)
        )

    if trainer_kwargs is None:
        trainer_kwargs = {}

    data_kwargs = {
        "num_workers": num_workers or 1,
        "batch_size": batch_size or 1,
    }
    if transform_kwargs:
        data_kwargs["transform_kwargs"] = transform_kwargs

    datamodule = fi.ImageClassificationData.from_fiftyone(
        predict_dataset=samples, **data_kwargs
    )
    embeddings = flash.Trainer(**trainer_kwargs).predict(
        model, datamodule=datamodule
    )
    embeddings = np.stack(sum(embeddings, []))

    if embeddings_field is not None:
        samples.set_values(embeddings_field, embeddings)
        return

    return embeddings


def _get_output(model, confidence_thresh, store_logits):
    if isinstance(model, fi.ImageClassifier):
        prev_args = {}
        try:
            prev_args = dict(inspect.getmembers(model.output))
        except AttributeError as e:
            pass

        kwargs = {
            "multi_label": prev_args.get("multi_label", False),
            "store_logits": store_logits,
        }

        if "threshold" in prev_args:
            kwargs["threshold"] = prev_args["threshold"]

        if confidence_thresh is not None:
            kwargs["threshold"] = confidence_thresh

        return fc.FiftyOneLabelsOutput(**kwargs)

    if isinstance(model, fi.ObjectDetector):
        return fdo.FiftyOneDetectionLabelsOutput(threshold=confidence_thresh)

    if isinstance(model, fi.SemanticSegmentation):
        return fso.FiftyOneSegmentationLabelsOutput()

    raise ValueError(
        "Unsupported model type %s. Supported model types are %s"
        % (type(model), _SUPPORTED_MODELS)
    )


_MODEL_TO_DATAMODULE_MAP = {
    fi.ObjectDetector: fi.ObjectDetectionData,
    fi.ImageClassifier: fi.ImageClassificationData,
    fi.SemanticSegmentation: fi.SemanticSegmentationData,
    fv.VideoClassifier: fv.VideoClassificationData,
}



"""
Plotly-powered view plots.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import itertools
from operator import itemgetter

import numpy as np

import eta.core.utils as etau

import plotly.subplots as ps
import plotly.graph_objects as go

import fiftyone.core.aggregations as foa

from .base import ViewPlot
from .plotly import PlotlyWidgetMixin


_DEFAULT_LAYOUT = dict(
    template="ggplot2", margin={"r": 0, "t": 30, "l": 0, "b": 0}
)

_DEFAULT_MARKER_COLOR = "#FF6D04"


class PlotlyViewPlot(PlotlyWidgetMixin, ViewPlot):
    """Base class for :class:`ViewPlot` instances with Plotly backends.

    Args:
        widget: a :class:`plotly:plotly.graph_objects.FigureWidget`
        init_view (None): an optional initial
            :class:`fiftyone.core.collections.SampleCollection` to load
    """

    def __init__(self, widget, init_view=None):
        self._traces = widget.data

        ViewPlot.__init__(self)
        PlotlyWidgetMixin.__init__(self, widget)  # must be last

        self.init_view = init_view
        if init_view is not None:
            self.connect()
            self.update_view(init_view)

    def _get_trace_updates(self, view, agg_results=None):
        raise NotImplementedError(
            "Subclass must implement _get_trace_updates()"
        )

    def _update_view(self, view, agg_results=None):
        updates = self._get_trace_updates(view, agg_results=agg_results)

        with self._widget.batch_update():
            for trace, update in zip(self._traces, updates):
                trace.update(update)

    def show(self, **kwargs):
        """Shows this plot.

        Args:
            **kwargs: optional parameters for
                :meth:`plotly:plotly.graph_objects.Figure.update_layout`
        """
        super().show(**kwargs)


class ViewGrid(PlotlyViewPlot):
    """A grid of :class:`PlotlyViewPlot` instances.

    Args:
        plots: a :class:`PlotlyViewPlot` or iterable of :class:`PlotlyViewPlot`
            instances
        shape (None): the ``(rows, cols)`` shape to use for the grid
        hgap (None): a horizontal spacing between the subplots, in ``[0, 1]``
        vgap (None): a vertical spacing between the subplots, in ``[0, 1]``
        init_view (None): an optional initial
            :class:`fiftyone.core.collections.SampleCollection` to load
        **kwargs: optional parameters for
            :meth:`plotly:plotly.graph_objects.Figure.update_layout`
    """

    def __init__(
        self, plots, shape=None, hgap=None, vgap=None, init_view=None, **kwargs
    ):
        if not etau.is_container(plots):
            plots = [plots]

        bad_plots = [p for p in plots if not isinstance(p, PlotlyViewPlot)]
        if bad_plots:
            raise ValueError(
                "Found %d unsupported plots; this class only supports %s plots"
                % (len(bad_plots), PlotlyViewPlot)
            )

        self.plots = plots
        self.shape = shape
        self.hgap = hgap
        self.vgap = vgap
        self.layout = kwargs

        self._aggregations = None
        self._agg_lengths = None

        self._init_aggregations()

        widget = self._make_widget()

        super().__init__(widget, init_view=init_view)

    def _init_aggregations(self):
        aggregations = []
        agg_lengths = []
        for plot in self.plots:
            aggs = plot._get_aggregations()
            if aggs is not None:
                aggregations.extend(aggs)
                agg_lengths.append(len(aggs))
            else:
                agg_lengths.append(None)

        if not aggregations:
            aggregations = None
            agg_lengths = None

        self._aggregations = aggregations
        self._agg_lengths = agg_lengths

    def _get_aggregations(self):
        return self._aggregations

    def _make_widget(self):
        widget = _make_composite_widget(
            self.plots, shape=self.shape, hgap=self.hgap, vgap=self.vgap
        )

        layout = _DEFAULT_LAYOUT.copy()
        layout.update(self.layout)

        widget.update_layout(**layout)

        return widget

    def _get_trace_updates(self, view, agg_results=None):
        if agg_results is None:
            agg_results = itertools.repeat(None)
        else:
            agg_iter = iter(agg_results)
            agg_results = []
            for agg_len in self._agg_lengths:
                if agg_len is None:
                    _agg_results = None
                else:
                    _agg_results = [next(agg_iter) for _ in range(agg_len)]

                agg_results.append(_agg_results)

        updates = []
        for plot, _agg_results in zip(self.plots, agg_results):
            update = plot._get_trace_updates(view, agg_results=_agg_results)
            updates.extend(update)

        return updates

    def _reopen(self):
        self._widget = self._make_widget()
        self._traces = self._widget.data


class CategoricalHistogram(PlotlyViewPlot):
    """A histogram of a categorial field.

    Args:
        field_or_expr: a field name, ``embedded.field.name``,
            :class:`fiftyone.core.expressions.ViewExpression`, or
            `MongoDB expression <https://docs.mongodb.com/manual/meta/aggregation-quick-reference/#aggregation-expressions>`_
            defining the field or expression to plot
        expr (None): an optional
            :class:`fiftyone.core.expressions.ViewExpression` or
            `MongoDB expression <https://docs.mongodb.com/manual/meta/aggregation-quick-reference/#aggregation-expressions>`_
            to apply to ``field_or_expr`` (which must be a field) before
            plotting
        order ("alphabetical"): the x-axis ordering strategy to use. Can be
            "alphabetical" to sort by field value, or "frequency" to sort in
            descending order of frequency, or a function suitable for
            ``sorted(items, key=order)``, where ``items`` is a list of
            ``(value, count)`` tuples
        xlabel (None): an optional x-label for the plot
        log (False): whether to use a log scale y-axis
        bargap (None): relative spacing between bars in ``[0, 1]``
        color (None): a color for the bars. Can be any color supported by
            :meth:`plotly:plotly.graph_objects.bar.Marker.color`
        opacity (None): an optional opacity for the bars in ``[0, 1]``
        init_view (None): an optional initial
            :class:`fiftyone.core.collections.SampleCollection` to load
        **kwargs: optional parameters for
            :meth:`plotly:plotly.graph_objects.Figure.update_layout`
    """

    def __init__(
        self,
        field_or_expr,
        expr=None,
        order="alphabetical",
        xlabel=None,
        log=None,
        bargap=None,
        color=None,
        opacity=None,
        init_view=None,
        **kwargs,
    ):
        self.field_or_expr = field_or_expr
        self.expr = expr
        self.order = order
        self.xlabel = xlabel
        self.log = log
        self.bargap = bargap
        self.color = color
        self.opacity = opacity
        self.layout = kwargs

        if order == "alphabetical":
            self._order = itemgetter(0)
            self._reverse = False
        elif order == "frequency":
            self._order = itemgetter(1)
            self._reverse = True
        else:
            self._order = order
            self._reverse = False

        self._figure = self._make_histogram()

        self._aggregations = [foa.CountValues(field_or_expr, expr=expr)]

        widget = self._make_widget()

        super().__init__(widget, init_view=init_view)

    def _make_widget(self):
        return go.FigureWidget(self._figure)

    def _make_histogram(self):
        if etau.is_str(self.field_or_expr):
            _field = self.field_or_expr.rsplit(".", 1)[-1]
        else:
            _field = "value"

        hover_lines = [
            "<b>%s: %%{x}</b>" % _field,
            "count: %{y}",
        ]
        hovertemplate = "<br>".join(hover_lines) + "<extra></extra>"

        if self.color is None:
            marker_color = _DEFAULT_MARKER_COLOR
        else:
            marker_color = self.color

        bar = go.Bar(
            hovertemplate=hovertemplate,
            marker_color=marker_color,
            opacity=self.opacity,
        )

        if self.xlabel is not None:
            xaxis_title = self.xlabel
        elif etau.is_str(self.field_or_expr):
            xaxis_title = self.field_or_expr
        else:
            xaxis_title = None

        layout = _DEFAULT_LAYOUT.copy()

        layout.update(
            dict(
                xaxis_title=xaxis_title,
                yaxis_title="count",
                bargap=self.bargap,
            )
        )

        if self.log:
            layout.update(dict(yaxis_type="log"))

        layout.update(self.layout)

        figure = go.Figure(bar)
        figure.update_layout(**layout)

        return figure

    def _get_aggregations(self):
        return self._aggregations

    def _get_trace_updates(self, view, agg_results=None):
        if view is not None:
            if agg_results is not None:
                counts = agg_results[0]
            else:
                counts = view.aggregate(self._aggregations[0])
        else:
            counts = None

        if counts:
            keys, values = zip(
                *sorted(counts.items(), key=self._order, reverse=self._reverse)
            )
        else:
            keys, values = None, None

        return [dict(x=keys, y=values)]

    def _reopen(self):
        self._widget = self._make_widget()
        self._traces = self._widget.data


class NumericalHistogram(PlotlyViewPlot):
    """A histogram of a numerical field.

    Args:
        field_or_expr: a field name, ``embedded.field.name``,
            :class:`fiftyone.core.expressions.ViewExpression`, or
            `MongoDB expression <https://docs.mongodb.com/manual/meta/aggregation-quick-reference/#aggregation-expressions>`_
            defining the field or expression to plot
        expr (None): an optional
            :class:`fiftyone.core.expressions.ViewExpression` or
            `MongoDB expression <https://docs.mongodb.com/manual/meta/aggregation-quick-reference/#aggregation-expressions>`_
            to apply to ``field_or_expr`` (which must be a field) before
            plotting
        bins (None): can be either an integer number of bins to generate or a
            monotonically increasing sequence specifying the bin edges to use.
            By default, 10 bins are created. If ``bins`` is an integer and no
            ``range`` is specified, bin edges are automatically computed from
            the bounds of the field
        range (None): a ``(lower, upper)`` tuple specifying a range in which to
            generate equal-width bins. Only applicable when ``bins`` is an
            integer or ``None``
        xlabel (None): an optional x-label for the plot
        log (False): whether to use a log scale y-axis
        color (None): a color for the bars. Can be any color supported by
            :meth:`plotly:plotly.graph_objects.bar.Marker.color`
        opacity (None): an optional opacity for the bars in ``[0, 1]``
        init_view (None): an optional initial
            :class:`fiftyone.core.collections.SampleCollection` to load
        **kwargs: optional parameters for
            :meth:`plotly:plotly.graph_objects.Figure.update_layout`
    """

    def __init__(
        self,
        field_or_expr,
        expr=None,
        bins=None,
        range=None,
        xlabel=None,
        log=None,
        color=None,
        opacity=None,
        init_view=None,
        **kwargs,
    ):
        self.field_or_expr = field_or_expr
        self.expr = expr
        self.bins = bins
        self.range = range
        self.xlabel = xlabel
        self.log = log
        self.color = color
        self.opacity = opacity
        self.layout = kwargs

        self._figure = self._make_histogram()

        self._aggregations = [
            foa.HistogramValues(
                field_or_expr, expr=expr, bins=bins, range=range
            )
        ]

        widget = self._make_widget()

        super().__init__(widget, init_view=init_view)

    def _make_widget(self):
        return go.FigureWidget(self._figure)

    def _make_histogram(self):
        if etau.is_str(self.field_or_expr):
            _field = self.field_or_expr.rsplit(".", 1)[-1]
        else:
            _field = "value"

        hover_lines = [
            "<b>count: %{y}</b>",
            "%s: [%%{customdata[0]:.2f}, %%{customdata[1]:.2f}]" % _field,
        ]
        hovertemplate = "<br>".join(hover_lines) + "<extra></extra>"

        if self.color is None:
            marker_color = _DEFAULT_MARKER_COLOR
        else:
            marker_color = self.color

        bar = go.Bar(
            offset=0,
            hovertemplate=hovertemplate,
            marker_color=marker_color,
            opacity=self.opacity,
        )

        if self.xlabel is not None:
            xaxis_title = self.xlabel
        elif etau.is_str(self.field_or_expr):
            xaxis_title = self.field_or_expr
        else:
            xaxis_title = None

        layout = _DEFAULT_LAYOUT.copy()

        layout.update(dict(xaxis_title=xaxis_title, yaxis_title="count"))

        if self.log:
            layout.update(dict(yaxis_type="log"))

        layout.update(self.layout)

        figure = go.Figure(bar)
        figure.update_layout(**layout)

        return figure

    def _get_aggregations(self):
        return self._aggregations

    def _get_trace_updates(self, view, agg_results=None):
        if view is not None:
            if agg_results is not None:
                counts, edges, _ = agg_results[0]
            else:
                counts, edges, _ = view.aggregate(self._aggregations[0])

            counts = np.asarray(counts)
            edges = np.asarray(edges)

            x = edges[:-1]
            y = counts
            width = edges[1:] - edges[:-1]
            customdata = np.stack((edges[:-1], edges[1:]), axis=1)
        else:
            x, y, width, customdata = None, None, None, None

        return [dict(x=x, y=y, width=width, customdata=customdata)]

    def _reopen(self):
        self._widget = self._make_widget()
        self._traces = self._widget.data


def _make_composite_widget(plots, shape=None, hgap=None, vgap=None):
    if shape is None:
        num_plots = len(plots)
        rows = int(np.floor(np.sqrt(num_plots)))
        cols = int(np.ceil(num_plots / rows))
    else:
        rows, cols = shape

    figure = ps.make_subplots(
        rows=rows, cols=cols, horizontal_spacing=hgap, vertical_spacing=vgap
    )

    num_traces = []
    for idx, plot in enumerate(plots):
        r, c = np.unravel_index(idx, (rows, cols))
        row = r + 1
        col = c + 1
        traces = plot._widget.data
        num_traces.append(len(traces))
        for trace in traces:
            figure.add_trace(trace, row=row, col=col)

        figure.update_xaxes(plot._widget.layout.xaxis, row=row, col=col)
        figure.update_yaxes(plot._widget.layout.yaxis, row=row, col=col)

    figure.update_layout(showlegend=False)

    return go.FigureWidget(figure)



"""
Utilities for working with datasets in
`VOC format <http://host.robots.ox.ac.uk/pascal/VOC>`_.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging
import os

import jinja2

import eta.core.utils as etau

import fiftyone.constants as foc
import fiftyone.core.labels as fol
import fiftyone.core.metadata as fom
import fiftyone.core.utils as fou
import fiftyone.utils.data as foud


logger = logging.getLogger(__name__)


class VOCDetectionDatasetImporter(
    foud.LabeledImageDatasetImporter, foud.ImportPathsMixin
):
    """Importer for VOC detection datasets stored on disk.

    See :ref:`this page <VOCDetectionDataset-import>` for format details.

    Args:
        dataset_dir (None): the dataset directory. If omitted, ``data_path``
            and/or ``labels_path`` must be provided
        data_path (None): an optional parameter that enables explicit control
            over the location of the media. Can be any of the following:

            -   a folder name like ``"data"`` or ``"data/"`` specifying a
                subfolder of ``dataset_dir`` where the media files reside
            -   an absolute directory path where the media files reside. In
                this case, the ``dataset_dir`` has no effect on the location of
                the data
            -   a filename like ``"data.json"`` specifying the filename of the
                JSON data manifest file in ``dataset_dir``
            -   an absolute filepath specifying the location of the JSON data
                manifest. In this case, ``dataset_dir`` has no effect on the
                location of the data
            -   a dict mapping filenames to absolute filepaths

            If None, this parameter will default to whichever of ``data/`` or
            ``data.json`` exists in the dataset directory
        labels_path (None): an optional parameter that enables explicit control
            over the location of the labels. Can be any of the following:

            -   a folder name like ``"labels"`` or ``"labels/"`` specifying the
                location of the labels in ``dataset_dir``
            -   an absolute folder path to the labels. In this case,
                ``dataset_dir`` has no effect on the location of the labels

            If None, the parameter will default to ``labels/``
        include_all_data (False): whether to generate samples for all images in
            the data directory (True) rather than only creating samples for
            images with label entries (False)
        extra_attrs (True): whether to load extra annotation attributes onto
            the imported labels. Supported values are:

            -   ``True``: load all extra attributes found
            -   ``False``: do not load extra attributes
            -   a name or list of names of specific attributes to load
        shuffle (False): whether to randomly shuffle the order in which the
            samples are imported
        seed (None): a random seed to use when shuffling
        max_samples (None): a maximum number of samples to import. By default,
            all samples are imported
    """

    def __init__(
        self,
        dataset_dir=None,
        data_path=None,
        labels_path=None,
        include_all_data=False,
        extra_attrs=True,
        shuffle=False,
        seed=None,
        max_samples=None,
    ):
        if dataset_dir is None and data_path is None and labels_path is None:
            raise ValueError(
                "At least one of `dataset_dir`, `data_path`, and "
                "`labels_path` must be provided"
            )

        data_path = self._parse_data_path(
            dataset_dir=dataset_dir,
            data_path=data_path,
            default="data/",
        )

        labels_path = self._parse_labels_path(
            dataset_dir=dataset_dir,
            labels_path=labels_path,
            default="labels/",
        )

        super().__init__(
            dataset_dir=dataset_dir,
            shuffle=shuffle,
            seed=seed,
            max_samples=max_samples,
        )

        self.data_path = data_path
        self.labels_path = labels_path
        self.include_all_data = include_all_data
        self.extra_attrs = extra_attrs

        self._image_paths_map = None
        self._labels_paths_map = None
        self._uuids = None
        self._iter_uuids = None
        self._num_samples = None

    def __iter__(self):
        self._iter_uuids = iter(self._uuids)
        return self

    def __len__(self):
        return self._num_samples

    def __next__(self):
        uuid = next(self._iter_uuids)

        labels_path = self._labels_paths_map.get(uuid, None)
        if labels_path:
            # Labeled image
            annotation = load_voc_detection_annotations(labels_path)

            # Use image filename from annotation file if possible
            if annotation.filename:
                _uuid = os.path.splitext(annotation.filename)[0]
            elif annotation.path:
                _uuid = os.path.splitext(os.path.basename(annotation.path))[0]
            else:
                _uuid = None

            if _uuid is not None:
                _uuid = fou.normpath(_uuid)

            if _uuid not in self._image_paths_map:
                _uuid = uuid

            if uuid in self._image_paths_map:
                image_path = self._image_paths_map[uuid]
            elif annotation.path and os.path.isfile(annotation.path):
                image_path = annotation.path
            else:
                raise ValueError("No image found for sample '%s'" % _uuid)

            image_metadata = annotation.metadata
            detections = annotation.to_detections(extra_attrs=self.extra_attrs)
        else:
            # Unlabeled image
            image_path = self._image_paths_map[uuid]
            image_metadata = None
            detections = None

        return image_path, image_metadata, detections

    @property
    def has_dataset_info(self):
        return False

    @property
    def has_image_metadata(self):
        return True

    @property
    def label_cls(self):
        return fol.Detections

    def setup(self):
        image_paths_map = self._load_data_map(
            self.data_path, ignore_exts=True, recursive=True
        )

        if self.labels_path is not None and os.path.isdir(self.labels_path):
            labels_path = fou.normpath(self.labels_path)
            labels_paths_map = {
                os.path.splitext(p)[0]: os.path.join(labels_path, p)
                for p in etau.list_files(labels_path, recursive=True)
                if etau.has_extension(p, ".xml")
            }
        else:
            labels_paths_map = {}

        uuids = set(labels_paths_map.keys())

        if self.include_all_data:
            uuids.update(image_paths_map.keys())

        uuids = self._preprocess_list(sorted(uuids))

        self._image_paths_map = image_paths_map
        self._labels_paths_map = labels_paths_map
        self._uuids = uuids
        self._num_samples = len(uuids)


class VOCDetectionDatasetExporter(
    foud.LabeledImageDatasetExporter, foud.ExportPathsMixin
):
    """Exporter that writes VOC detection datasets to disk.

    See :ref:`this page <VOCDetectionDataset-export>` for format details.

    Args:
        export_dir (None): the directory to write the export. This has no
            effect if ``data_path`` and ``labels_path`` are absolute paths
        data_path (None): an optional parameter that enables explicit control
            over the location of the exported media. Can be any of the
            following:

            -   a folder name like ``"data"`` or ``"data/"`` specifying a
                subfolder of ``export_dir`` in which to export the media
            -   an absolute directory path in which to export the media. In
                this case, the ``export_dir`` has no effect on the location of
                the data
            -   a JSON filename like ``"data.json"`` specifying the filename of
                the manifest file in ``export_dir`` generated when
                ``export_media`` is ``"manifest"``
            -   an absolute filepath specifying the location to write the JSON
                manifest file when ``export_media`` is ``"manifest"``. In this
                case, ``export_dir`` has no effect on the location of the data

            If None, the default value of this parameter will be chosen based
            on the value of the ``export_media`` parameter
        labels_path (None): an optional parameter that enables explicit control
            over the location of the exported labels. Can be any of the
            following:

            -   a folder name like ``"labels"`` or ``"labels/"`` specifying the
                location in ``export_dir`` in which to export the labels
            -   an absolute folder path to which to export the labels. In this
                case, the ``export_dir`` has no effect on the location of the
                labels

            If None, the labels will be exported into ``export_dir`` using the
            default folder name
        export_media (None): controls how to export the raw media. The
            supported values are:

            -   ``True``: copy all media files into the output directory
            -   ``False``: don't export media
            -   ``"move"``: move all media files into the output directory
            -   ``"symlink"``: create symlinks to the media files in the output
                directory
            -   ``"manifest"``: create a ``data.json`` in the output directory
                that maps UUIDs used in the labels files to the filepaths of
                the source media, rather than exporting the actual media

            If None, the default value of this parameter will be chosen based
            on the value of the ``data_path`` parameter
        rel_dir (None): an optional relative directory to strip from each input
            filepath to generate a unique identifier for each image. When
            exporting media, this identifier is joined with ``data_path`` and
            ``labels_path`` to generate output paths for each exported image
            and labels file. This argument allows for populating nested
            subdirectories that match the shape of the input paths. The path is
            converted to an absolute path (if necessary) via
            :func:`fiftyone.core.utils.normalize_path`
        include_paths (True): whether to include the absolute paths to the
            images in the ``<path>`` elements of the exported XML
        image_format (None): the image format to use when writing in-memory
            images to disk. By default, ``fiftyone.config.default_image_ext``
            is used
        extra_attrs (True): whether to include extra object attributes in the
            exported labels. Supported values are:

            -   ``True``: export all extra attributes found
            -   ``False``: do not export extra attributes
            -   a name or list of names of specific attributes to export
    """

    def __init__(
        self,
        export_dir=None,
        data_path=None,
        labels_path=None,
        export_media=None,
        rel_dir=None,
        include_paths=True,
        image_format=None,
        extra_attrs=True,
    ):
        data_path, export_media = self._parse_data_path(
            export_dir=export_dir,
            data_path=data_path,
            export_media=export_media,
            default="data/",
        )

        labels_path = self._parse_labels_path(
            export_dir=export_dir,
            labels_path=labels_path,
            default="labels/",
        )

        super().__init__(export_dir=export_dir)

        self.data_path = data_path
        self.labels_path = labels_path
        self.export_media = export_media
        self.rel_dir = rel_dir
        self.include_paths = include_paths
        self.image_format = image_format
        self.extra_attrs = extra_attrs

        self._writer = None
        self._media_exporter = None

    @property
    def requires_image_metadata(self):
        return True

    @property
    def label_cls(self):
        return fol.Detections

    def setup(self):
        self._writer = VOCAnnotationWriter()
        self._media_exporter = foud.ImageExporter(
            self.export_media,
            export_path=self.data_path,
            rel_dir=self.rel_dir,
            default_ext=self.image_format,
        )
        self._media_exporter.setup()

        etau.ensure_dir(self.labels_path)

    def export_sample(self, image_or_path, detections, metadata=None):
        out_image_path, uuid = self._media_exporter.export(image_or_path)

        if detections is None:
            return

        out_labels_path = os.path.join(
            self.labels_path, os.path.splitext(uuid)[0] + ".xml"
        )

        if metadata is None:
            metadata = fom.ImageMetadata.build_for(image_or_path)

        if self.include_paths:
            path = out_image_path
        else:
            path = None

        annotation = VOCAnnotation.from_labeled_image(
            metadata,
            detections,
            path=path,
            filename=uuid,
            extra_attrs=self.extra_attrs,
        )
        self._writer.write(annotation, out_labels_path)

    def close(self, *args):
        self._media_exporter.close()


class VOCAnnotation(object):
    """Class representing a VOC annotations file.

    Args:
        path (None): the path to the image
        folder (None): the name of the folder containing the image
        filename (None): the image filename
        segmented (None): whether the objects are segmented
        metadata (None): a :class:`fiftyone.core.metadata.ImageMetadata`
            instance
        objects (None): a list of :class:`VOCObject` instances
    """

    def __init__(
        self,
        path=None,
        folder=None,
        filename=None,
        segmented=None,
        metadata=None,
        objects=None,
    ):
        if filename is None and path:
            filename = os.path.basename(path)

        self.path = path
        self.folder = folder
        self.filename = filename
        self.segmented = segmented
        self.metadata = metadata
        self.objects = objects or []

    def to_detections(self, extra_attrs=True):
        """Returns a :class:`fiftyone.core.labels.Detections` representation of
        the objects in the annotation.

        Args:
            extra_attrs (True): whether to load extra annotation attributes
                onto the imported labels. Supported values are:

                -   ``True``: load all extra attributes found
                -   ``False``: do not load extra attributes
                -   a name or list of names of specific attributes to load

        Returns:
            a :class:`fiftyone.core.labels.Detections`
        """
        if self.metadata is None:
            raise ValueError(
                "Must have metadata in order to convert to `Detections` format"
            )

        frame_size = (self.metadata.width, self.metadata.height)
        return fol.Detections(
            detections=[
                obj.to_detection(frame_size, extra_attrs=extra_attrs)
                for obj in self.objects
            ]
        )

    @classmethod
    def from_labeled_image(
        cls, metadata, detections, path=None, filename=None, extra_attrs=True
    ):
        """Creates a :class:`VOCAnnotation` instance for the given labeled
        image data.

        Args:
            metadata: a :class:`fiftyone.core.metadata.ImageMetadata` instance
                for the image
            detections: a :class:`fiftyone.core.labels.Detections`
            path (None): the absolute path to the image
            filename (None): the filename of the image
            extra_attrs (True): whether to include extra object attributes.
                Supported values are:

                -   ``True``: include all extra attributes found
                -   ``False``: do not include extra attributes
                -   a name or list of names of specific attributes to include

        Returns:
            a :class:`VOCAnnotation`
        """
        objects = []
        for detection in detections.detections:
            obj = VOCObject.from_detection(
                detection, metadata, extra_attrs=extra_attrs
            )
            objects.append(obj)

        return cls(
            path=path, filename=filename, metadata=metadata, objects=objects
        )

    @classmethod
    def from_xml(cls, xml_path):
        """Creates a :class:`VOCAnnotation` instance from an XML annotations
        file.

        Args:
            xml_path: the path to the XML file

        Returns:
            a :class:`VOCAnnotation`
        """
        d = fou.load_xml_as_json_dict(xml_path)
        return cls.from_dict(d)

    @classmethod
    def from_dict(cls, d):
        """Creates a :class:`VOCAnnotation` instance from a JSON dict
        representation.

        Args:
            d: a JSON dict

        Returns:
            a :class:`VOCAnnotation`
        """
        annotation = d["annotation"]

        path = annotation.get("path", None)
        folder = annotation.get("folder", None)
        filename = annotation.get("filename", None)
        segmented = annotation.get("segmented", None)

        if "size" in annotation:
            size = annotation["size"]
            width = size.get("width", None)
            height = size.get("height", None)
            depth = size.get("depth", None)

            metadata = fom.ImageMetadata(
                width=int(width) if width else None,
                height=int(height) if height else None,
                num_channels=int(depth) if depth else None,
            )
        else:
            metadata = None

        _objects = _ensure_list(annotation.get("object", []))
        objects = [VOCObject.from_annotation_dict(do) for do in _objects]

        return cls(
            path=path,
            folder=folder,
            filename=filename,
            segmented=segmented,
            metadata=metadata,
            objects=objects,
        )


class VOCObject(object):
    """An object in VOC detection format.

    Args:
        name: the object label
        bndbox: a :class:`VOCBoundingBox`
        **attributes: additional custom attributes
    """

    def __init__(self, name, bndbox, **attributes):
        self.name = name
        self.bndbox = bndbox
        self.attributes = attributes

    @classmethod
    def from_annotation_dict(cls, d):
        """Creates a :class:`VOCObject` from a VOC annotation dict.

        Args:
            d: an annotation dict

        Returns:
            a :class:`VOCObject`
        """
        name = d["name"]
        bndbox = VOCBoundingBox.from_bndbox_dict(d["bndbox"])

        # Handles CVAT exported attributes
        if "attributes" in d:
            cvat_attrs = d.pop("attributes", {}).pop("attribute", {})
            cvat_attrs = _ensure_list(cvat_attrs)
            cvat_attrs = {a["name"]: a["value"] for a in cvat_attrs}
            d.update(cvat_attrs)

        attributes = {
            k: _parse_attribute(d[k])
            for k in set(d.keys()) - {"name", "bndbox"}
        }
        return cls(name, bndbox, **attributes)

    @classmethod
    def from_detection(cls, detection, metadata, extra_attrs=True):
        """Creates a :class:`VOCObject` from a
        :class:`fiftyone.core.labels.Detection`.

        Args:
            detection: a :class:`fiftyone.core.labels.Detection`
            metadata: a :class:`fiftyone.core.metadata.ImageMetadata` instance
                for the image
            extra_attrs (True): whether to include extra object attributes.
                Supported values are:

                -   ``True``: include all extra attributes found
                -   ``False``: do not include extra attributes
                -   a name or list of names of specific attributes to include

        Returns:
            a :class:`VOCObject`
        """
        name = detection.label

        frame_size = (metadata.width, metadata.height)
        bndbox = VOCBoundingBox.from_detection_format(
            detection.bounding_box, frame_size
        )

        attributes = _get_attributes(detection, extra_attrs)

        return cls(name, bndbox, **attributes)

    def to_detection(self, frame_size, extra_attrs=True):
        """Returns a :class:`fiftyone.core.labels.Detection` representation of
        the object.

        Args:
            frame_size: the ``(width, height)`` of the image
            extra_attrs (True): whether to include extra annotation attributes
                on the object. Supported values are:

                -   ``True``: include all extra attributes found
                -   ``False``: do not include extra attributes
                -   a name or list of names of specific attributes to include

        Returns:
            a :class:`fiftyone.core.labels.Detection`
        """
        label = self.name
        bounding_box = self.bndbox.to_detection_format(frame_size)

        if extra_attrs == True:
            attributes = self.attributes
        elif extra_attrs == False:
            attributes = {}
        else:
            if etau.is_str(extra_attrs):
                extra_attrs = [extra_attrs]

            attributes = {
                name: self.attributes.get(name, None) for name in extra_attrs
            }

        return fol.Detection(
            label=label, bounding_box=bounding_box, **attributes
        )


class VOCBoundingBox(object):
    """A bounding box in VOC detection format.

    Args:
        xmin: the top-left x coordinate
        ymin: the top-left y coordinate
        xmax: the bottom-right x coordinate
        ymax: the bottom-right y coordinate
    """

    def __init__(self, xmin, ymin, xmax, ymax):
        self.xmin = xmin
        self.ymin = ymin
        self.xmax = xmax
        self.ymax = ymax

    @classmethod
    def from_bndbox_dict(cls, d):
        """Creates a :class:`VOCBoundingBox` from a ``bndbox`` dict.

        Args:
            d: a ``bndbox`` dict

        Returns:
            a :class:`VOCBoundingBox`
        """
        return cls(
            int(float(d["xmin"])),
            int(float(d["ymin"])),
            int(float(d["xmax"])),
            int(float(d["ymax"])),
        )

    @classmethod
    def from_detection_format(cls, bounding_box, frame_size):
        """Creates a :class:`VOCBoundingBox` from a bounding box stored in
        :class:`fiftyone.core.labels.Detection` format.

        Args:
            bounding_box: ``[x-top-left, y-top-left, width, height]``
            frame_size: the ``(width, height)`` of the image

        Returns:
            a :class:`VOCBoundingBox`
        """
        x, y, w, h = bounding_box
        width, height = frame_size
        return cls(
            int(width * x),
            int(height * y),
            int(width * (x + w)),
            int(height * (y + h)),
        )

    def to_detection_format(self, frame_size):
        """Returns a representation of the bounding box suitable for storing in
        the ``bounding_box`` field of a
        :class:`fiftyone.core.labels.Detection`.

        Args:
            frame_size: the ``(width, height)`` of the image

        Returns:
            ``[x-top-left, y-top-left, width, height]``
        """
        width, height = frame_size
        x = self.xmin / width
        y = self.ymin / height
        w = (self.xmax - self.xmin) / width
        h = (self.ymax - self.ymin) / height
        return [x, y, w, h]


class VOCAnnotationWriter(object):
    """Class for writing annotations in VOC format.

    See :ref:`this page <VOCDetectionDataset-export>` for format details.
    """

    def __init__(self):
        environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(foc.RESOURCES_DIR),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        self.template = environment.get_template("voc_annotation_template.xml")

    def write(self, annotation, xml_path):
        """Writes the annotations to disk.

        Args:
            annotation: a :class:`VOCAnnotation` instance
            xml_path: the path to write the annotation XML file
        """
        if annotation.metadata is not None:
            metadata = annotation.metadata
        else:
            metadata = fom.ImageMetadata()

        xml_str = self.template.render(
            {
                "path": annotation.path,
                "filename": annotation.filename,
                "folder": annotation.folder,
                "width": metadata.width or "",
                "height": metadata.height or "",
                "depth": metadata.num_channels or "",
                "database": None,
                "segmented": annotation.segmented,
                "objects": annotation.objects,
            }
        )
        etau.write_file(xml_str, xml_path)


def load_voc_detection_annotations(xml_path):
    """Loads the VOC detection annotations from the given XML file.

    See :ref:`this page <VOCDetectionDataset-import>` for format details.

    Args:
        xml_path: the path to the annotations XML file

    Returns:
        a :class:`VOCAnnotation` instance
    """
    return VOCAnnotation.from_xml(xml_path)


VOC_DETECTION_CLASSES = [
    "aeroplane",
    "bicycle",
    "bird",
    "boat",
    "bottle",
    "bus",
    "car",
    "cat",
    "chair",
    "cow",
    "diningtable",
    "dog",
    "horse",
    "motorbike",
    "person",
    "pottedplant",
    "sheep",
    "sofa",
    "train",
    "tvmonitor",
]


def _ensure_list(value):
    if value is None:
        return []

    if isinstance(value, list):
        return value

    return [value]


def _get_attributes(label, extra_attrs):
    if extra_attrs == True:
        return dict(label.iter_attributes())

    if extra_attrs == False:
        return {}

    if etau.is_str(extra_attrs):
        extra_attrs = [extra_attrs]

    return {
        name: label.get_attribute_value(name, None) for name in extra_attrs
    }


def _parse_attribute(value):
    try:
        return int(value)
    except:
        pass

    try:
        return float(value)
    except:
        pass

    if etau.is_str(value):
        if value in ("True", "true"):
            return True

        if value in ("False", "false"):
            return False

        if value in ("None", ""):
            return None

    return value



"""
FiftyOne v0.15.0 admin revision.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging


logger = logging.getLogger(__name__)


def up(db):
    db = db["fiftyone"]

    dataset_names = [d.get("name", None) for d in db.datasets.find({})]

    colls_in_use = set()
    for dataset_name in dataset_names:
        dataset_dict = db.datasets.find_one({"name": dataset_name})
        sample_coll_name = dataset_dict.get("sample_collection_name", None)
        if sample_coll_name:
            colls_in_use.add(sample_coll_name)
            colls_in_use.add("frames." + sample_coll_name)

    # Only collections with these prefixes may be deleted
    prefixes = ("samples.", "frames.", "patches.", "clips.")

    drop_colls = []
    for coll_name in db.list_collection_names():
        if coll_name not in colls_in_use and any(
            coll_name.startswith(prefix) for prefix in prefixes
        ):
            drop_colls.append(coll_name)

    if drop_colls:
        logger.info(
            "Dropping %d orphan collections that were unintentionally left "
            "behind when datasets were deleted",
            len(drop_colls),
        )
        for coll_name in drop_colls:
            db.drop_collection(coll_name)


def down(db):
    pass
